"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import logging
import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, Union

from geedim import utils

logger = logging.getLogger(__name__)
root_stac_url = 'https://earthengine-stac.storage.googleapis.com/catalog/catalog.json'


class StacItem:

    def __init__(self, name: str, item_dict: Dict):
        """
        Image/collection STAC container class.  Provides access to band properties and root property descriptions.

        Parameters
        ----------
        name: str
            The image/collection ID/name
        item_dict: dict
            The raw STAC dict for the image/collection.
        """
        self._name = name
        self._item_dict = item_dict
        self._descriptions = self._get_descriptions(item_dict)
        self._band_props = self._get_band_props(item_dict)

    def _get_descriptions(self, item_dict: Dict) -> Union[Dict[str, str], None]:
        """ Return a dictionary with property names as keys, and descriptions as values. """
        if 'summaries' not in item_dict:
            return None
        if 'gee:schema' in item_dict['summaries']:
            gee_schema = item_dict['summaries']['gee:schema']
        elif 'gee:collection_schema' in item_dict['summaries']:
            gee_schema = item_dict['summaries']['gee:collection_schema']
        else:
            return None
        descriptions = {item['name']: item['description'] for item in gee_schema}
        return descriptions

    def _get_band_props(self, item_dict: Dict) -> Union[Dict[str, Dict], None]:
        """
        Return a dictionary of band properties, with band names as keys, and properties as values.
        """
        if not ('summaries' in item_dict and 'eo:bands' in item_dict['summaries']):
            logger.warning(f'There is no STAC band information for {self._name}')
            return None

        summaries = item_dict['summaries']
        ee_band_props = summaries['eo:bands']
        # if the gsd is the same across all bands, there is a `gsd` key in summaries, otherwise there are `gsd` keys
        # for each item in ee_band_props
        global_gsd = summaries['gsd'] if 'gsd' in summaries else None
        band_props = {}
        # a list of the EE band properties we want to copy
        prop_keys = [
            'name', 'description', 'center_wavelength', 'gee:wavelength', 'gee:units', 'gee:scale', 'gee:offset'
        ]

        def strip_gee(key: str):
            """ Remove 'gee:' from the start of `key` if it is there. """
            return key[4:] if key.startswith('gee:') else key

        for ee_band_dict in ee_band_props:
            band_dict = {
                strip_gee(prop_key): ee_band_dict[prop_key]
                for prop_key in prop_keys
                if prop_key in ee_band_dict
            }  # yapf: disable
            gsd = ee_band_dict['gsd'] if 'gsd' in ee_band_dict else global_gsd
            gsd = gsd[0] if isinstance(gsd, (list, tuple)) else gsd
            band_dict.update(gsd=gsd) if gsd else None
            band_props[ee_band_dict['name']] = band_dict
        return band_props

    @property
    def name(self) -> str:
        """ ID/name of the contained image/collection STAC. """
        return self._name

    @property
    def descriptions(self) -> Union[Dict[str, str], None]:
        """ Dictionary of property descriptions with property names as keys, and descriptions as values. """
        return self._descriptions

    @property
    def band_props(self) -> Union[Dict[str, Dict], None]:
        """ Dictionary of band properties, with band names as keys, and properties as values. """
        return self._band_props

    @property
    def license(self) -> Union[str, None]:
        """ Terms of use / license. """
        url = None
        if 'links' in self._item_dict:
            for link in self._item_dict['links']:
                if ('rel' in link) and (link['rel'] == 'license') and ('href' in link):
                    url = link['href']
                    break
        return url


@utils.singleton
class StacCatalog:

    def __init__(self):
        """ Singleton class to interface to the EE STAC, and retrieve image/collection STAC data. """
        self._filename = utils.root_path.joinpath('geedim/data/ee_stac_urls.json')
        self._session = utils.retry_session()
        self._url_dict = None
        self._cache = {}
        self._lock = threading.Lock()

    @property
    def url_dict(self) -> Dict[str, str]:
        """ Dictionary with image/collection IDs/names as keys, and STAC URLs as values. """
        if not self._url_dict:
            # delay reading the json file until it is needed.
            with open(self._filename, 'r') as f:
                self._url_dict = json.load(f)
        return self._url_dict

    def _traverse_stac(self, url: str, url_dict: Dict) -> Dict:
        """
        Recursive & threaded EE STAC tree traversal that returns the `url_dict` i.e. a dict with image/collection
        IDs/names as keys, and the corresponding json STAC URLs as values.
        """
        response = self._session.get(url)
        if not response.ok:
            logger.warning(f'Error reading {url}: ' + str(response.content))
            return url_dict
        response_dict = response.json()
        if 'type' in response_dict:
            if (response_dict['type'].lower() == 'collection'):
                # we have reached a leaf node
                if (('gee:type' in response_dict) and
                    (response_dict['gee:type'].lower() in ['image_collection', 'image'])):
                    # we have reached an image / image collection leaf node
                    with self._lock:
                        url_dict[response_dict['id']] = url
                        logger.debug(f'ID: {response_dict["id"]}, Type: {response_dict["gee:type"]}, URL: {url}')
                return url_dict

            with ThreadPoolExecutor() as executor:
                # traverse the sub-tree links in a thread pool
                futures = []
                for link in response_dict['links']:
                    if link['rel'].lower() == 'child':
                        futures.append(executor.submit(self._traverse_stac, link['href'], url_dict))
                for future in as_completed(futures):
                    url_dict = future.result()
        return url_dict

    def refresh_url_dict(self):
        """ Update `url_dict` with the latest from EE STAC. """
        url_dict = {}
        url_dict = self._traverse_stac(root_stac_url, url_dict)
        self._url_dict = dict(sorted(url_dict.items()))

    def write_url_dict(self, filename=None):
        """ Write the ``url_dict`` to file. """
        if filename is None:
            filename = self._filename
        with open(filename, 'w') as f:
            json.dump(self.url_dict, f)

    def get_item_dict(self, name: str):
        """
        Get the raw STAC dict for a given an image/collection name/ID.

        Parameters
        ----------
        name: str
            ID/name of the image/collection whose STAC data to retrieve.

        Returns
        -------
        dict
            Image/collection STAC data in a dict, if it exists, otherwise None.
        """
        coll_name = utils.split_id(name)[0]
        if coll_name in self.url_dict:
            name = coll_name

        # store item dicts in a cache so we don't have to request them more than once
        if name not in self._cache:
            if name not in self.url_dict:
                logger.warning(f'There is no STAC entry for: {name}')
                self._cache[name] = None
            else:
                response = self._session.get(self.url_dict[name])
                self._cache[name] = response.json()
        return self._cache[name]

    def get_item(self, name: str) -> StacItem:
        """
        Get a STAC container instance for a given an image/collection name/ID.

        Parameters
        ----------
        name: str
            The ID/name of the image/collection whose STAC container to retrieve.

        Returns
        -------
        StacItem
            image/collection STAC container, if it exists, otherwise None.
        """
        coll_name = utils.split_id(name)[0]
        if coll_name in self.url_dict:
            name = coll_name
        item_dict = self.get_item_dict(name)
        return StacItem(name, item_dict) if item_dict else None



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import time
from typing import Dict

import ee
import pytest
from geedim import MaskedImage
from geedim.enums import ResamplingMethod
from geedim.utils import split_id, get_projection, get_bounds, Spinner, resample, asset_id
from rasterio.features import bounds

from .conftest import get_image_std


@pytest.mark.parametrize('id, exp_split', [('A/B/C', ('A/B', 'C')), ('ABC', ('', 'ABC')), (None, (None, None))])
def test_split_id(id, exp_split):
    """ Test split_id(). """
    assert split_id(id) == exp_split


def test_get_bounds(const_image_25ha_file, region_25ha):
    """ Test get_bounds(). """
    raster_bounds = bounds(get_bounds(const_image_25ha_file, expand=0))
    test_bounds = bounds(region_25ha)
    assert raster_bounds == pytest.approx(test_bounds, abs=.001)


def test_get_projection(s2_sr_masked_image):
    """ Test get_projection().  """
    min_proj = get_projection(s2_sr_masked_image.ee_image, min_scale=True)
    min_crs = min_proj.crs().getInfo()
    min_scale = min_proj.nominalScale().getInfo()
    max_proj = get_projection(s2_sr_masked_image.ee_image, min_scale=False)
    max_crs = max_proj.crs().getInfo()
    max_scale = max_proj.nominalScale().getInfo()

    assert min_crs.startswith('EPSG:')
    assert min_crs == max_crs
    assert max_scale == 60
    assert min_scale == 10


def test_spinner():
    """ Test Spinner class. """
    spinner = Spinner(label='test', interval=0.1)
    assert not spinner.is_alive()
    with spinner:
        assert spinner._run
        assert spinner.is_alive()
        time.sleep(0.5)
    assert not spinner._run
    assert not spinner.is_alive()


# yapf: disable
@pytest.mark.parametrize(
    'image_id, method, std_scale', [
        ('l9_image_id', ResamplingMethod.bilinear, 30),
        ('s2_sr_image_id', ResamplingMethod.average, 60),
        ('modis_nbar_image_id', ResamplingMethod.bicubic, 500),
    ]
)
# yapf: enable
def test_resample_fixed(
    image_id: str, method: ResamplingMethod, std_scale: float, region_10000ha: Dict, request: pytest.FixtureRequest
):
    """ Test that resample() smooths images with a fixed projection. """
    image_id = request.getfixturevalue(image_id)
    before_image = ee.Image(image_id)
    after_image = resample(before_image, method)

    assert (
        get_image_std(after_image, region_10000ha, std_scale) < get_image_std(before_image, region_10000ha, std_scale)
    )


@pytest.mark.parametrize(
    'masked_image, method, std_scale', [
        ('user_masked_image', ResamplingMethod.bilinear, 100),
        ('landsat_ndvi_masked_image', ResamplingMethod.average, 60)
    ]
)
def test_resample_comp(
    masked_image: str, method: ResamplingMethod, std_scale: float, region_10000ha: Dict, request: pytest.FixtureRequest
):
    """ Test that resample() leaves composite images unaltered. """
    masked_image: MaskedImage = request.getfixturevalue(masked_image)
    before_image = masked_image.ee_image
    after_image = resample(before_image, method)

    assert (
        get_image_std(after_image, region_10000ha, std_scale) == get_image_std(before_image, region_10000ha, std_scale)
    )


@pytest.mark.parametrize(
    'filename, folder, exp_id', [
        ('file', 'folder', 'projects/folder/assets/file'),
        ('fp1/fp2/fp3', 'folder', 'projects/folder/assets/fp1-fp2-fp3'),
        ('file', 'folder/sub-folder', 'projects/folder/assets/sub-folder/file'),
        ('file', None, 'file'),
        ('projects/folder/assets/file', None, 'projects/folder/assets/file'),
    ]
)
def test_asset_id(filename: str, folder: str, exp_id: str):
    """ Test asset_id() works as expected. """
    id = asset_id(filename, folder)
    assert id == exp_id



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import pathlib
from datetime import datetime
from typing import Dict, Tuple, List
import copy

import ee
import numpy as np
import pytest
import rasterio as rio
from rasterio import features, warp
from rasterio.crs import CRS
from geedim.download import BaseImage
from geedim.enums import ResamplingMethod, ExportType
from geedim import utils
from rasterio import Affine, windows


class BaseImageLike:
    """ Emulate BaseImage for _get_tile_shape() and tiles(). """

    def __init__(
        self, shape: Tuple[int, int], count: int = 10, dtype: str = 'uint16', transform: Affine = Affine.identity()
    ):
        self.shape = shape
        self.count = count
        self.dtype = dtype
        self.transform = transform
        dtype_size = np.dtype(dtype).itemsize
        self.size = shape[0] * shape[1] * count * dtype_size

    _tiles = BaseImage._tiles
    _get_tile_shape = BaseImage._get_tile_shape


@pytest.fixture(scope='session')
def user_base_image() -> BaseImage:
    """ A BaseImage instance where the encapsulated image has no fixed projection or ID.  """
    return BaseImage(ee.Image([1, 2, 3]))


@pytest.fixture(scope='session')
def user_fix_base_image() -> BaseImage:
    """
    A BaseImage instance where the encapsulated image has a fixed projection (EPSG:4326), a scale in degrees,
    and no footprint or ID.
    """
    return BaseImage(ee.Image([1, 2, 3]).reproject(crs='EPSG:4326', scale=30))


@pytest.fixture(scope='session')
def s2_sr_base_image(s2_sr_image_id: str) -> BaseImage:
    """ A BaseImage instance encapsulating a Sentinel-2 image.  Covers `region_*ha`.  """
    return BaseImage.from_id(s2_sr_image_id)


@pytest.fixture(scope='session')
def l9_base_image(l9_image_id: str) -> BaseImage:
    """ A BaseImage instance encapsulating a Landsat-9 image.  Covers `region_*ha`.  """
    return BaseImage.from_id(l9_image_id)


@pytest.fixture(scope='session')
def l8_base_image(l8_image_id: str) -> BaseImage:
    """ A BaseImage instance encapsulating a Landsat-8 image.  Covers `region_*ha`.  """
    return BaseImage.from_id(l8_image_id)


@pytest.fixture(scope='session')
def landsat_ndvi_base_image(landsat_ndvi_image_id: str) -> BaseImage:
    """ A BaseImage instance encapsulating a Landsat NDVI composite image.  Covers `region_*ha`.  """
    return BaseImage.from_id(landsat_ndvi_image_id)


@pytest.fixture(scope='session')
def modis_nbar_base_image_unbnd(modis_nbar_image_id: str) -> BaseImage:
    """ A BaseImage instance encapsulating an unbounded MODIS NBAR image in its native CRS.  Covers `region_*ha`.  """
    return BaseImage(ee.Image(modis_nbar_image_id))

@pytest.fixture(scope='session')
def modis_nbar_base_image(modis_nbar_image_id: str, region_100ha: Dict) -> BaseImage:
    """ A BaseImage instance encapsulating a MODIS NBAR image.  Covers `region_*ha`.  """
    return BaseImage(ee.Image(modis_nbar_image_id).clip(region_100ha))


def bounds_polygon(left: float, bottom: float, right: float, top: float, crs: str = None):
    """ Return a geojson polygon of the given bounds. """
    coordinates = [[
        [left, bottom],
        [left, top],
        [right, top],
        [right, bottom],
        [left, bottom],
    ]]
    poly = dict(type='Polygon', coordinates=coordinates)
    if crs and crs != 'EPSG:4326':
        poly.update(crs=dict(type='name', properties=dict(name=crs)))
    return poly


def _test_export_profile(exp_profile, tgt_profile, transform_shape=False):
    """
    Test an export/download rasterio profile against a target rasterio profile.
    ``crs_transform`` specifies whether the export/download profile was generated by specifying crs_transform.
    """

    for key in ['dtype', 'count']:
        assert exp_profile[key] == tgt_profile[key]

    assert utils.rio_crs(exp_profile['crs']) == utils.rio_crs(tgt_profile['crs'])

    assert exp_profile['transform'][0] == tgt_profile['transform'][0]

    if transform_shape:
        for key in ['width', 'height']:
            assert exp_profile[key] == tgt_profile[key]
        assert exp_profile['transform'][:6] == pytest.approx(tgt_profile['transform'][:6], rel=1e-9)
    else:
        assert exp_profile['transform'][:6] ==  pytest.approx(tgt_profile['transform'][:6], rel=0.05)

    tgt_bounds = windows.bounds(
        windows.Window(0, 0, tgt_profile['width'], tgt_profile['height']), tgt_profile['transform']
    )
    exp_bounds = windows.bounds(
        windows.Window(0, 0, exp_profile['width'], exp_profile['height']), exp_profile['transform']
    )
    if transform_shape:
        assert exp_bounds == pytest.approx(tgt_bounds, rel=1e-9)
    else:
        assert exp_bounds == pytest.approx(tgt_bounds, rel=0.05)


def test_id_name(user_base_image: BaseImage, s2_sr_base_image: BaseImage):
    """ Test `id` and `name` properties for different scenarios. """
    assert user_base_image.id is None
    assert user_base_image.name is None
    # check that BaseImage.from_id() sets id without a getInfo
    assert (s2_sr_base_image._id is not None)
    assert s2_sr_base_image.name == s2_sr_base_image.id.replace('/', '-')


def test_user_props(user_base_image: BaseImage):
    """ Test non fixed projection image properties (other than id and has_fixed_projection). """
    assert user_base_image.crs is None
    assert user_base_image.scale is None
    assert user_base_image.transform is None
    assert user_base_image.shape is None
    assert user_base_image.date is None
    assert user_base_image.size is None
    assert user_base_image.footprint is None
    assert user_base_image.dtype == 'uint8'
    assert user_base_image.count == 3


def test_fix_user_props(user_fix_base_image: BaseImage):
    """ Test fixed projection image properties (other than id and has_fixed_projection). """
    assert user_fix_base_image.crs == 'EPSG:4326'
    assert user_fix_base_image.scale < 1
    assert user_fix_base_image.transform is not None
    assert user_fix_base_image.shape is None
    assert user_fix_base_image.date is None
    assert user_fix_base_image.size is None
    assert user_fix_base_image.footprint is None
    assert user_fix_base_image.dtype == 'uint8'
    assert user_fix_base_image.count == 3


def test_s2_props(s2_sr_base_image: BaseImage):
    """ Test fixed projection S2 image properties (other than id and has_fixed_projection). """
    min_band_info = s2_sr_base_image._ee_info['bands'][1]
    assert s2_sr_base_image.crs == min_band_info['crs']
    assert s2_sr_base_image.scale == min_band_info['crs_transform'][0]
    assert s2_sr_base_image.transform == Affine(*min_band_info['crs_transform'])
    assert s2_sr_base_image.shape == min_band_info['dimensions'][::-1]
    assert s2_sr_base_image.date == datetime.utcfromtimestamp(s2_sr_base_image.properties['system:time_start'] / 1000)
    assert s2_sr_base_image.size is not None
    assert s2_sr_base_image.footprint is not None
    assert s2_sr_base_image.footprint['type'] == 'Polygon'
    assert s2_sr_base_image.dtype == 'uint32'
    assert s2_sr_base_image.count == len(s2_sr_base_image._ee_info['bands'])


@pytest.mark.parametrize(
    'base_image', ['landsat_ndvi_base_image', 's2_sr_base_image', 'l9_base_image', 'modis_nbar_base_image']
)
def test_band_props(base_image: str, request: pytest.FixtureRequest):
    """ Test `band_properties` completeness for generic/user/reflectance images. """
    base_image: BaseImage = request.getfixturevalue(base_image)
    assert base_image.band_properties is not None
    assert [bd['name'] for bd in base_image.band_properties] == [bd['id'] for bd in base_image._ee_info['bands']]
    for key in ['gsd', 'description']:
        has_key = [key in bd for bd in base_image.band_properties]
        assert all(has_key)


def test_has_fixed_projection(user_base_image: BaseImage, user_fix_base_image: BaseImage, s2_sr_base_image):
    """ Test the `has_fixed_projection` property.  """
    assert not user_base_image.has_fixed_projection
    assert user_fix_base_image.has_fixed_projection
    assert s2_sr_base_image.has_fixed_projection


# yapf: disable
@pytest.mark.parametrize(
    'ee_data_type_list, exp_dtype', [
        ([{'precision': 'int', 'min': 10, 'max': 11}, {'precision': 'int', 'min': 100, 'max': 101}], 'uint8'),
        ([{'precision': 'int', 'min': -128, 'max': -100}, {'precision': 'int', 'min': 0, 'max': 127}], 'int8'),
        ([{'precision': 'int', 'min': 256, 'max': 257}], 'uint16'),
        ([{'precision': 'int', 'min': -32768, 'max': 32767}], 'int16'),
        ([{'precision': 'int', 'min': 2**15, 'max': 2**32 - 1}], 'uint32'),
        ([{'precision': 'int', 'min': -2**31, 'max': 2**31 - 1}], 'int32'),
        ([{'precision': 'float', 'min': 0., 'max': 1.e9}, {'precision': 'float', 'min': 0., 'max': 1.}], 'float32'),
        ([{'precision': 'int', 'min': 0., 'max': 2**31 - 1}, {'precision': 'float', 'min': 0., 'max': 1.}], 'float64'),
        ([{'precision': 'int', 'min': 0, 'max': 255}, {'precision': 'double', 'min': -1e100, 'max': 1e100}], 'float64'),
    ]
)
# yapf: enable
def test_min_dtype(ee_data_type_list: List, exp_dtype: str):
    """ Test BasicImage.__get_min_dtype() with emulated EE info dicts.  """
    ee_info = dict(bands=[])
    for ee_data_type in ee_data_type_list:
        ee_info['bands'].append(dict(data_type=ee_data_type))
    assert BaseImage._get_min_dtype(ee_info) == exp_dtype


def test_convert_dtype_error():
    """ Test BaseImage.test_convert_dtype() raises an error with incorrect dtype. """
    with pytest.raises(TypeError):
        BaseImage._convert_dtype(ee.Image(1), dtype='unknown')


@pytest.mark.parametrize('size, exp_str', [(1024, '1.02 KB'), (234.56e6, '234.56 MB'), (1e9, '1.00 GB')])
def test_str_format_size(size: int, exp_str: str):
    """ Test formatting of byte sizes as human readable strings. """
    assert BaseImage._str_format_size(size) == exp_str


@pytest.mark.parametrize(
    'params', [
        dict(),
        dict(crs='EPSG:4326'),
        dict(crs='EPSG:4326', region='region_25ha'),
        dict(crs='EPSG:4326', scale=100),
        dict(crs='EPSG:4326', crs_transform=Affine.identity()),
        dict(crs='EPSG:4326', shape=(100, 100)),
        dict(region='region_25ha', scale=100),
        dict(crs_transform=Affine.identity(), shape=(100, 100)),
    ]
)  # yapf: disable
def test_prepare_no_fixed_projection(user_base_image: BaseImage, params: Dict, request: pytest.FixtureRequest):
    """
    Test BaseImage._prepare_for_export raises an exception when the image has no fixed projection, and insufficient
    crs & region defining parameters are specified.
    """
    if 'region' in params:
        params['region'] = request.getfixturevalue(params['region'])
    with pytest.raises(ValueError) as ex:
        user_base_image._prepare_for_export(**params)
    assert 'does not have a fixed projection' in str(ex)


@pytest.mark.parametrize(
    'base_image, params', [
        ('user_fix_base_image', dict()),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326')),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326', scale=100)),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326', crs_transform=Affine.identity())),
        ('modis_nbar_base_image_unbnd', dict(crs='EPSG:4326', shape=(100, 100))),
        ('modis_nbar_base_image_unbnd', dict(crs_transform=Affine.identity(), shape=(100, 100))),
    ]
)  # yapf: disable
def test_prepare_unbounded(base_image: BaseImage, params: Dict, request: pytest.FixtureRequest):
    """
    Test BaseImage._prepare_for_export raises an exception when the image is unbounded, and insufficient bounds
    defining parameters are specified.
    """
    base_image: BaseImage = request.getfixturevalue(base_image)
    with pytest.raises(ValueError) as ex:
        base_image._prepare_for_export(**params)
    assert 'This image is unbounded' in str(ex)


def test_prepare_exceptions(user_base_image: BaseImage, user_fix_base_image: BaseImage, region_25ha: Dict):
    """ Test remaining BaseImage._prepare_for_export() error cases. """
    with pytest.raises(ValueError):
        # EPSG:4326 and no scale
        user_fix_base_image._prepare_for_export(region=region_25ha)
    with pytest.raises(ValueError):
        # no fixed projection and resample
        user_base_image._prepare_for_export(
            region=region_25ha, crs='EPSG:3857', scale=30, resampling=ResamplingMethod.bilinear
        )


@pytest.mark.parametrize(
    'base_image', ['s2_sr_base_image', 'l9_base_image', 'modis_nbar_base_image'],
)
def test_prepare_defaults(base_image: str, request: pytest.FixtureRequest):
    """ Test BaseImage._prepare_for_export() with no (i.e. default) arguments with bounded images. """
    base_image: BaseImage = request.getfixturevalue(base_image)
    exp_image = base_image._prepare_for_export()

    tgt_profile = base_image.profile
    _test_export_profile(exp_image.profile, tgt_profile, True)
    assert exp_image.scale == base_image.scale


@pytest.mark.parametrize(
    'base_image, param_image', [
        ('s2_sr_base_image', 's2_sr_base_image'),
        ('l9_base_image', 's2_sr_base_image'),
        ('modis_nbar_base_image', 's2_sr_base_image'),
        ('user_base_image', 's2_sr_base_image'),
        ('user_fix_base_image', 's2_sr_base_image'),
        ('l9_base_image', 'l9_base_image'),
        ('s2_sr_base_image', 'l9_base_image'),
        ('modis_nbar_base_image', 'l9_base_image'),
        ('user_base_image', 'l9_base_image'),
        ('user_fix_base_image', 'l9_base_image'),
        ('l9_base_image', 'modis_nbar_base_image'),
        ('s2_sr_base_image', 'modis_nbar_base_image'),
        ('modis_nbar_base_image', 'modis_nbar_base_image'),
        ('user_base_image', 'modis_nbar_base_image'),
        ('user_fix_base_image', 'modis_nbar_base_image'),
    ]  # yapf: disable
)
def test_prepare_transform_shape(base_image: str, param_image: str, request: pytest.FixtureRequest):
    """ Test BaseImage._prepare_for_export() with crs_transform and shape parameters. """
    base_image: BaseImage = request.getfixturevalue(base_image)
    param_image: BaseImage = request.getfixturevalue(param_image)

    exp_params = dict(
        crs=param_image.crs, crs_transform=param_image.transform, shape=param_image.shape, dtype=param_image.dtype
    )
    exp_image = base_image._prepare_for_export(**exp_params)

    tgt_profile = param_image.profile
    tgt_profile.update(count=base_image.count)

    _test_export_profile(exp_image.profile, tgt_profile, True)
    assert exp_image.scale == param_image.scale


@pytest.mark.parametrize(
    'base_image, param_image', [
        ('s2_sr_base_image', 's2_sr_base_image'),
        ('l9_base_image', 's2_sr_base_image'),
        ('modis_nbar_base_image', 's2_sr_base_image'),
        ('user_base_image', 's2_sr_base_image'),
        ('user_fix_base_image', 's2_sr_base_image'),
        ('l9_base_image', 'l9_base_image'),
        ('s2_sr_base_image', 'l9_base_image'),
        ('modis_nbar_base_image', 'l9_base_image'),
        ('user_base_image', 'l9_base_image'),
        ('user_fix_base_image', 'l9_base_image'),
        ('l9_base_image', 'modis_nbar_base_image'),
        ('s2_sr_base_image', 'modis_nbar_base_image'),
        ('modis_nbar_base_image', 'modis_nbar_base_image'),
        ('user_base_image', 'modis_nbar_base_image'),
        ('user_fix_base_image', 'modis_nbar_base_image'),
    ]
)  # yapf: disable
def test_prepare_region_scale(base_image: str, param_image: str, region_25ha: dict, request: pytest.FixtureRequest):
    """ Test BaseImage._prepare_for_export() with region and scale parameters. """
    base_image: BaseImage = request.getfixturevalue(base_image)
    _param_image: BaseImage = request.getfixturevalue(param_image)
    param_image = copy.deepcopy(_param_image)   # avoid changing session fixture
    # clip param_image, so that param_image.transform can be tested against below
    param_image.ee_image = param_image.ee_image.clip(region_25ha)

    exp_params = dict(
        crs=param_image.crs, region=param_image.footprint, scale=param_image.scale, dtype=param_image.dtype
    )
    exp_image = base_image._prepare_for_export(**exp_params)

    tgt_profile = param_image.profile
    tgt_profile.update(count=base_image.count)

    _test_export_profile(exp_image.profile, tgt_profile, base_image.id == param_image.id)
    assert exp_image.scale == param_image.scale

    # test export bounds contain target bounds
    exp_bounds = features.bounds(exp_image.footprint)
    tgt_bounds = features.bounds(param_image.footprint)
    tgt_footprint_crs = (
        param_image.footprint['crs']['properties']['name'] if 'crs' in param_image.footprint else 'EPSG:4326'
    )
    exp_crs = CRS.from_string(utils.rio_crs(exp_image.crs))
    tgt_bounds = warp.transform_bounds(tgt_footprint_crs, exp_crs, *tgt_bounds)
    assert (
        (exp_bounds[0] <= tgt_bounds[0]) and (exp_bounds[1] <= tgt_bounds[1]) and
        (exp_bounds[2] >= tgt_bounds[2]) and (exp_bounds[3] >= tgt_bounds[3])
    )


@pytest.mark.parametrize(
    'base_image, bands', [
        ('s2_sr_base_image', ['B1', 'B5']),
        ('l9_base_image', ['SR_B4', 'SR_B3', 'SR_B2']),
    ]
)  # yapf: disable
def test_prepare_bands(base_image: str, bands: List[str], region_25ha: dict, request: pytest.FixtureRequest):
    """ Test BaseImage._prepare_for_export() with bands parameter. """
    base_image: BaseImage = request.getfixturevalue(base_image)
    param_image = BaseImage(base_image.ee_image.select(bands))

    exp_image = base_image._prepare_for_export(bands=bands)

    assert exp_image.count == len(bands)
    for attr in ['crs', 'transform', 'scale', 'shape', 'band_properties']:
        assert exp_image.__getattribute__(attr) == param_image.__getattribute__(attr)


def test_prepare_bands_error(s2_sr_base_image):
    """ Test BaseImage._prepare_for_export() raises an error with incorrect bands. """
    with pytest.raises(ValueError):
        s2_sr_base_image._prepare_for_export(bands=['unknown'])


@pytest.mark.parametrize(
    'base_image', ['s2_sr_base_image', 'l9_base_image', 'modis_nbar_base_image'],
)
def test_prepare_for_download(
    base_image: str, request: pytest.FixtureRequest
):
    """ Test BaseImage._prepare_for_download() sets rasterio profile as expected.  """
    base_image: BaseImage = request.getfixturevalue(base_image)
    exp_image, exp_profile = base_image._prepare_for_download()
    tgt_profile = base_image.profile
    _test_export_profile(exp_profile, tgt_profile, True)
    assert exp_profile['nodata'] is not None


@pytest.mark.parametrize(
    'dtype, exp_nodata', [
        ('uint8', 0), ('int8', -2**7), ('uint16', 0), ('int16', -2**15), ('uint32', 0), ('int32', -2**31),
        ('float32', float('nan')), ('float64', float('nan')),
    ]
)
def test_prepare_nodata(user_fix_base_image: BaseImage, region_25ha: Dict, dtype: str, exp_nodata: float):
    """ Test BaseImage._prepare_for_download() sets rasterio profile nodata correctly for different dtypes.  """
    exp_image, exp_profile = user_fix_base_image._prepare_for_download(region=region_25ha, scale=30, dtype=dtype)
    assert exp_image.dtype == dtype
    if np.isnan(exp_profile['nodata']):
        assert np.isnan(exp_nodata)
    else:
        assert exp_profile['nodata'] == exp_nodata


@pytest.mark.parametrize(
    'src_image, dtype', [
        ('s2_sr_base_image', 'float32'),
        ('l9_base_image', None),
        ('modis_nbar_base_image', None),
    ]
)  # yapf: disable
def test_scale_offset(src_image: str, dtype: str, region_100ha: Dict, request: pytest.FixtureRequest):
    """ Test BaseImage._prepare_for_export(scale_offset=True) gives expected properties and reflectance ranges. """

    src_image: BaseImage = request.getfixturevalue(src_image)
    exp_image = src_image._prepare_for_export(scale_offset=True)
    assert exp_image.crs == src_image.crs
    assert exp_image.scale == src_image.scale
    assert exp_image.band_properties == src_image.band_properties
    assert exp_image.dtype == dtype or 'float64'

    def get_min_max_refl(base_image: BaseImage) -> Dict:
        """ Get the min & max of each reflectance band of base_image.  """
        band_props = base_image.band_properties
        refl_bands = [
            bp['name'] for bp in band_props if ('center_wavelength' in bp) and (bp['center_wavelength'] < 1)
        ]  # yapf: disable
        ee_image = base_image.ee_image.select(refl_bands)
        min_max_dict = ee_image.reduceRegion(
            reducer=ee.Reducer.minMax(), geometry=region_100ha, bestEffort=True
        ).getInfo() # yapf: disable
        min_dict = {k: v for k, v in min_max_dict.items() if 'min' in k}
        max_dict = {k: v for k, v in min_max_dict.items() if 'max' in k}
        return min_dict, max_dict

    # test the scaled and offset reflectance values lie between -0.5 and 1.5
    exp_min, exp_max = get_min_max_refl(exp_image)
    assert all(np.array(list(exp_min.values())) >= -0.5)
    assert all(np.array(list(exp_max.values())) <= 1.5)


def test_tile_shape():
    """ Test BaseImage._get_tile_shape() satisfies the tile size limit for different image shapes. """
    max_tile_dim = 10000

    for max_tile_size in range(4, 32, 4):
        for height in range(1, 11000, 100):
            for width in range(1, 11000, 100):
                exp_shape = (height, width)
                exp_image = BaseImageLike(shape=exp_shape)  # emulate a BaseImage
                tile_shape, num_tiles = exp_image._get_tile_shape(
                    max_tile_size=max_tile_size, max_tile_dim=max_tile_dim
                )
                assert all(np.array(tile_shape) <= np.array(exp_shape))
                assert all(np.array(tile_shape) <= max_tile_dim)
                tile_image = BaseImageLike(shape=tile_shape)
                assert tile_image.size <= (max_tile_size << 20)


@pytest.mark.parametrize(
    'image_shape, tile_shape, image_transform', [
        ((1000, 500), (101, 101), Affine.identity()),
        ((1000, 100), (101, 101), Affine.scale(1.23)),
        ((1000, 102), (101, 101), Affine.scale(1.23) * Affine.translation(12, 34)),
    ]
)  # yapf: disable
def test_tiles(image_shape: Tuple, tile_shape: Tuple, image_transform: Affine):
    """ Test continuity and coverage of tiles. """
    exp_image = BaseImageLike(shape=image_shape, transform=image_transform)
    tiles = [tile for tile in exp_image._tiles(tile_shape=tile_shape)]

    # test window coverage, and window & transform continuity
    prev_tile = tiles[0]
    accum_window = prev_tile.window
    for tile in tiles[1:]:
        accum_window = windows.union(accum_window, tile.window)
        assert all(np.array(tile._shape) <= np.array(tile_shape))

        if tile.window.row_off == prev_tile.window.row_off:
            assert tile.window.col_off == (prev_tile.window.col_off + prev_tile.window.width)
            assert tile._transform == pytest.approx(
                (prev_tile._transform * Affine.translation(prev_tile.window.width, 0)), rel=0.001
            )
        else:
            assert tile.window.row_off == (prev_tile.window.row_off + prev_tile.window.height)
            assert tile._transform == pytest.approx(
                prev_tile._transform * Affine.translation(-prev_tile.window.col_off, prev_tile.window.height), rel=0.001
            )
        prev_tile = tile
    assert (accum_window.height, accum_window.width) == exp_image.shape


def test_download_transform_shape(user_base_image: str, tmp_path: pathlib.Path, request: pytest.FixtureRequest):
    """ Test download file properties and pixel data with crs_transform and shape arguments. """
    tgt_prof = dict(crs='EPSG:3857', transform=Affine(1, 0, 0, 0, -1, 0), width=10, height=10, dtype='uint16', count=3)

    # form download parameters from tgt_prof
    shape = (tgt_prof['height'], tgt_prof['width'])
    tgt_bounds = windows.bounds(
        windows.Window(0, 0, *shape[::-1]), tgt_prof['transform']
    )
    download_params = dict(
        crs=tgt_prof['crs'], crs_transform=tgt_prof['transform'], shape=shape, dtype=tgt_prof['dtype']
    )
    filename = tmp_path.joinpath('test.tif')
    user_base_image.download(filename, **download_params)
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        _test_export_profile(ds.profile, tgt_prof, 'crs_transform' in download_params)
        array = ds.read()
        for i in range(ds.count):
            assert np.all(array[i] == i + 1)


def test_download_region_scale(user_base_image: str, tmp_path: pathlib.Path, request: pytest.FixtureRequest):
    """ Test download file properties and pixel data with region and scale arguments. """
    tgt_prof = dict(crs='EPSG:3857', transform=Affine(1, 0, 0, 0, -1, 0), width=10, height=10, dtype='uint16', count=3)

    # form download parameters from tgt_prof
    shape = (tgt_prof['height'], tgt_prof['width'])
    tgt_bounds = windows.bounds(
        windows.Window(0, 0, *shape[::-1]), tgt_prof['transform']
    )
    region = bounds_polygon(*tgt_bounds, crs=tgt_prof['crs'])
    download_params = dict(crs=tgt_prof['crs'], region=region, scale=tgt_prof['transform'][0], dtype=tgt_prof['dtype'])
    filename = tmp_path.joinpath('test.tif')
    user_base_image.download(filename, **download_params)
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        _test_export_profile(ds.profile, tgt_prof, 'crs_transform' in download_params)
        array = ds.read()
        for i in range(ds.count):
            assert np.all(array[i] == i + 1)


def test_overviews(user_base_image: BaseImage, region_25ha: Dict, tmp_path: pathlib.Path):
    """ Test overviews get built on download. """
    filename = tmp_path.joinpath('test_user_download.tif')
    user_base_image.download(filename, region=region_25ha, crs='EPSG:3857', scale=1)
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        for band_i in range(ds.count):
            assert len(ds.overviews(band_i + 1)) > 0
            assert ds.overviews(band_i + 1)[0] == 2


def test_metadata(s2_sr_base_image: BaseImage, region_25ha: Dict, tmp_path: pathlib.Path):
    """ Test metadata is written to a downloaded file. """
    filename = tmp_path.joinpath('test_s2_band_subset_download.tif')
    s2_sr_base_image.download(filename, region=region_25ha, crs='EPSG:3857', scale=60, bands=['B9'])
    assert filename.exists()
    with rio.open(filename, 'r') as ds:
        assert 'LICENSE' in ds.tags()
        assert len(ds.tags()['LICENSE']) > 0
        assert 'B9' in ds.descriptions
        band_dict = ds.tags(1)
        for key in ['gsd', 'name', 'description']:
            assert key in band_dict


@pytest.mark.parametrize('type', [ExportType.drive, ExportType.asset, ExportType.cloud])  # yapf: disable
def test_start_export(type, user_fix_base_image: BaseImage, region_25ha: Dict):
    """ Test start of a small export. """
    # Note that this export should start successfully, but will ultimately fail for the asset and cloud options.  For
    # the asset option, there will be an overwrite issue.  For cloud storage, there is no 'geedim' bucket.
    task = user_fix_base_image.export(
        'test_export', type=type, folder='geedim', scale=30, region=region_25ha, wait=False
    )
    assert task.active()
    assert task.status()['state'] == 'READY'


def __test_export_asset(user_fix_base_image: BaseImage, region_25ha: Dict):
    """ Test start of a small export to EE asset. """
    # TODO: consider removing this slow test in favour of the equivalent integration test
    filename = f'test_export_{np.random.randint(1<<31)}'
    folder = 'geedim'
    asset_id = utils.asset_id(filename, folder)
    # Note: to allow parallel tests exporting to assets, we use random asset names to prevent conflicts.  The test
    # must wait for export to complete and clean up after itself, otherwise lots of test assets will accumulate in my
    # geedim cloud project.

    try:
        # export and test asset exists
        task = user_fix_base_image.export(
            filename, type=ExportType.asset, folder=folder, scale=30, region=region_25ha, wait=True
        )
        assert task.status()['state'] == 'COMPLETED'
        assert ee.data.getAsset(asset_id) is not None
    finally:
        # delete asset
        try:
            ee.data.deleteAsset(asset_id)
        except ee.ee_exception.EEException:
            pass


def test_download_bigtiff(s2_sr_base_image: BaseImage, tmp_path: pathlib.Path):
    """ Test that BIGTIFF gets set in the profile of images larger than 4GB. """
    exp_image, profile = s2_sr_base_image._prepare_for_download()
    assert exp_image.size >= 4e9
    assert 'bigtiff' in profile
    assert profile['bigtiff']


def test_prepare_ee_geom(l9_base_image: BaseImage, tmp_path: pathlib.Path):
    """ Test that _prepare_for_export works with an ee.Geometry region at native crs and scale (Issue #6). """
    region = l9_base_image.ee_image.geometry()
    exp_image = l9_base_image._prepare_for_export(region=region)
    assert exp_image.scale == l9_base_image.scale


@pytest.mark.parametrize(
    'base_image, exp_value', [
        ('s2_sr_base_image', True),
        ('l9_base_image', True),
        ('modis_nbar_base_image_unbnd', False),
        ('user_base_image', False),
        ('user_fix_base_image', False),
    ]
)  # yapf: disable
def test_bounded(base_image: str, exp_value: bool, request: pytest.FixtureRequest):
    """ Test BaseImage.bounded has correct value for different images. """
    base_image: BaseImage = request.getfixturevalue(base_image)
    assert base_image.bounded == exp_value


# TODO:
# - export(): test an export of small file
# - different generic collection images are downloaded ok (perhaps this goes with MaskedImage more than BaseImage)
# - test float mask/nodata in downloaded image
# - test mult tile download has no discontinuities

##



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from datetime import datetime
from typing import List, Union, Dict

import ee
import numpy as np
import pytest
from geedim import schema
from geedim.collection import MaskedCollection
from geedim.enums import CompositeMethod, ResamplingMethod
from geedim.errors import UnfilteredError, InputImageError
from geedim.mask import MaskedImage
from geedim.utils import split_id, get_projection

from .conftest import get_image_std


@pytest.fixture()
def l4_5_image_list(l4_image_id, l5_masked_image) -> List[Union[str, MaskedImage]]:
    """ A list of landsat 4 & 5 image ID / MaskedImage's """
    return [l4_image_id, l5_masked_image]


@pytest.fixture()
def l8_9_image_list(l8_image_id, l9_masked_image) -> List[Union[str, MaskedImage]]:
    """ A list of landsat 8 & 9 image IDs/ MaskedImage's """
    return [l8_image_id, l9_masked_image]


@pytest.fixture()
def s2_sr_image_list() -> List[Union[str, MaskedImage]]:
    """ A list of Sentinel-2 SR image IDs/ MaskedImage's """
    return [
        'COPERNICUS/S2_SR/20211004T080801_20211004T083709_T34HEJ',
        'COPERNICUS/S2_SR/20211123T081241_20211123T083704_T34HEJ',
        MaskedImage.from_id('COPERNICUS/S2_SR/20220107T081229_20220107T083059_T34HEJ')
    ]


@pytest.fixture()
def gedi_image_list() -> List[Union[str, MaskedImage]]:
    """ A list of GEDI canopy top height IDs/ MaskedImage's """
    return [
        'LARSE/GEDI/GEDI02_A_002_MONTHLY/202008_018E_036S', 'LARSE/GEDI/GEDI02_A_002_MONTHLY/202009_018E_036S',
        MaskedImage.from_id('LARSE/GEDI/GEDI02_A_002_MONTHLY/202005_018E_036S')
    ]


def test_split_id():
    """ Test split_id(). """
    coll_name, im_id = split_id('A/B/C')
    assert coll_name == 'A/B'
    assert im_id == 'C'
    coll_name, im_id = split_id('ABC')
    assert coll_name == ''
    assert im_id == 'ABC'


@pytest.mark.parametrize('name', ['l9_image_id', 'gch_image_id'])
def test_from_name(name: str, request):
    """ Test MaskedCollection.from_name() for non Sentinel-2 collections. """
    name = request.getfixturevalue(name)
    name, _ = split_id(name)
    gd_collection = MaskedCollection.from_name(name)
    assert gd_collection._name == name
    assert gd_collection.schema is not None
    assert len(gd_collection.schema) >= len(schema.default_prop_schema)
    assert gd_collection.ee_collection == ee.ImageCollection(name)


@pytest.mark.parametrize('name', ['s2_sr_image_id', 's2_toa_image_id'])
def test_from_name_s2(name: str, request):
    """
    Test MaskedCollection.from_name() filters out images that don't have matching cloud probability (for Sentinel-2
    collections).
    """
    name = request.getfixturevalue(name)
    name, _ = split_id(name)
    gd_collection = MaskedCollection.from_name(name)
    assert gd_collection._name == name
    assert gd_collection.schema is not None
    assert gd_collection.schema == schema.s2_prop_schema
    # check ee_collection is not the full unfiltered collection
    assert gd_collection.ee_collection != ee.ImageCollection(name)
    # check one of the problem images is not in the collection
    # 20220305T075809_20220305T082125_T35HKD, 0220122T081241_20220122T083135_T34HEJ,
    # 20220226T080909_20220226T083100_T34HEH are other options
    filt = ee.Filter.eq('system:index', '0220122T081241_20220122T083135_T34HEJ')
    filt_collection = gd_collection.ee_collection.filter(filt)
    assert filt_collection.size().getInfo() == 0


def test_unfiltered_error(s2_sr_image_id):
    """ Test UnfilteredError is raised when calling `properties` or `composite` on an unfiltered collection. """
    gd_collection = MaskedCollection.from_name(split_id(s2_sr_image_id)[0])
    with pytest.raises(UnfilteredError):
        _ = gd_collection.properties
    with pytest.raises(UnfilteredError):
        _ = gd_collection.composite()


def test_from_list_errors(landsat_image_ids, s2_image_ids, user_masked_image):
    """ Test MaskedCollection.from_list() various error cases. """
    with pytest.raises(InputImageError):
        # test an error is raised when an image has no 'id'/'system:time_start' property
        MaskedCollection.from_list([landsat_image_ids[0], user_masked_image])

    with pytest.raises(InputImageError):
        # test an error is raised when images are not from compatible collections
        MaskedCollection.from_list([*landsat_image_ids])

    with pytest.raises(TypeError):
        # test an error is raised when an image type is not recognised
        MaskedCollection.from_list([landsat_image_ids[0], {}])

    with pytest.raises(ValueError):
        # test an error is raised when image list is empty
        MaskedCollection.from_list([])


@pytest.mark.parametrize('image_list', ['l4_5_image_list', 'l8_9_image_list'])
def test_from_list_landsat(image_list: str, request):
    """
    Test MaskedCollection.from_list() works with landsat images from different, but spectrally compatible
    collections.
    """
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    assert 'LANDSAT' in gd_collection.name
    assert gd_collection.properties is not None


@pytest.mark.parametrize('image_list', ['s2_sr_image_list', 'gedi_image_list'])
def test_from_list(image_list: str, request):
    """
    Test MaskedCollection.from_list() generates a valid MaskedCollection object from lists of cloud/shadow
    maskable, and generic images.
    """
    image_list: List = request.getfixturevalue(image_list)
    image_ids = [im_obj if isinstance(im_obj, str) else im_obj.id for im_obj in image_list]
    gd_collection = MaskedCollection.from_list(image_list)
    assert gd_collection.properties is not None
    assert gd_collection.schema is not None
    assert len(gd_collection.schema) >= len(schema.default_prop_schema)
    assert gd_collection.name == split_id(image_ids[0])[0]
    assert len(gd_collection.properties) == len(image_list)
    assert list(gd_collection.properties.keys()) == image_ids
    assert set(gd_collection.schema.keys()) > set(list(gd_collection.properties.values())[0].keys())
    assert gd_collection.properties_table is not None
    assert gd_collection.schema_table is not None


@pytest.mark.parametrize('image_list', ['s2_sr_image_list', 'gedi_image_list'])
def test_from_list_order(image_list: str, request):
    """ Test MaskedCollection.from_list() maintains the order of the provided image list. """
    image_list: List = request.getfixturevalue(image_list)[::-1]
    image_ids = [im_obj if isinstance(im_obj, str) else im_obj.id for im_obj in image_list]
    gd_collection = MaskedCollection.from_list(image_list)
    assert list(gd_collection.properties.keys()) == image_ids


def test_from_list_ee_image(gedi_image_list: List):
    """ Test MaskedCollection.from_list() with an ee.Image in the list. """
    image_ids = [im_obj if isinstance(im_obj, str) else im_obj.id for im_obj in gedi_image_list]
    image_list = gedi_image_list
    image_list[1] = ee.Image(image_list[1])
    gd_collection = MaskedCollection.from_list(image_list)
    assert list(gd_collection.properties.keys()) == image_ids

@pytest.mark.parametrize(
    'image_list, add_props', [
        ('s2_sr_image_list', ['AOT_RETRIEVAL_ACCURACY', 'CLOUDY_PIXEL_PERCENTAGE']),
        ('l8_9_image_list', ['CLOUD_COVER', 'GEOMETRIC_RMSE_VERIFY'])
    ]
)  # yapf: disable
def test_from_list_add_props(image_list: str, add_props: List, request: pytest.FixtureRequest):
    """
    Test MaskedCollection.from_list(add_props=...) contains the add_props in properties and schema.
    """
    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list, add_props=add_props)
    assert gd_collection.properties is not None
    assert gd_collection.schema is not None
    assert len(gd_collection.schema) > len(schema.default_prop_schema)
    assert all([add_prop in gd_collection.schema.keys() for add_prop in add_props])
    assert all([gd_collection.schema[add_prop]['abbrev'] is not None for add_prop in add_props])
    assert all([add_prop in list(gd_collection.properties.values())[0].keys() for add_prop in add_props])
    assert gd_collection.properties_table is not None
    assert gd_collection.schema_table is not None


@pytest.mark.parametrize(
    'name, start_date, end_date, region, fill_portion, cloudless_portion, is_csmask', [
        ('LANDSAT/LC09/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha', 0, 50, True),
        ('LANDSAT/LE07/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha', 0, 0, True),
        ('LANDSAT/LT05/C02/T1_L2', '2005-01-01', '2006-02-01', 'region_100ha', 40, 50, True),
        ('COPERNICUS/S2_SR', '2022-01-01', '2022-01-15', 'region_100ha', 0, 50, True),
        ('COPERNICUS/S2_HARMONIZED', '2022-01-01', '2022-01-15', 'region_100ha', 50, 40, True),
        ('LARSE/GEDI/GEDI02_A_002_MONTHLY', '2021-08-01', '2021-09-01', 'region_100ha', .1, 0, False)
    ]
)
def test_search(
    name, start_date: str, end_date: str, region: str, fill_portion: float, cloudless_portion: float, is_csmask, request
):
    """
    Test MaskedCollection.search() with fill / cloudless portion filters gives valid results for different cloud/shadow
    maskable, and generic collections.
    """
    region: Dict = request.getfixturevalue(region)
    gd_collection = MaskedCollection.from_name(name)
    searched_collection = gd_collection.search(
        start_date, end_date, region, fill_portion=fill_portion, cloudless_portion=cloudless_portion
    )

    properties = searched_collection.properties
    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')
    im_dates = np.array(
        [datetime.utcfromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    # test FILL_PORTION in expected range
    im_fill_portions = np.array([im_props['FILL_PORTION'] for im_props in properties.values()])
    assert np.all(im_fill_portions >= fill_portion) and np.all(im_fill_portions <= 100)
    if is_csmask:  # is a cloud/shadow masked collection
        # test CLOUDLESS_PORTION in expected range
        im_cl_portions = np.array([im_props['CLOUDLESS_PORTION'] for im_props in properties.values()])
        assert np.all(im_cl_portions >= cloudless_portion) and np.all(im_cl_portions <= 100)
    # test search result image dates lie between `start_date` and `end_date`
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)
    assert set(searched_collection.schema.keys()) == set(list(properties.values())[0].keys())
    # test search result image dates are sorted
    assert np.all(sorted(im_dates) == im_dates)


def test_empty_search(region_100ha):
    """ Test MaskedCollection.search() for empty search results. """
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    searched_collection = gd_collection.search('2022-01-01', '2022-01-02', region_100ha, cloudless_portion=100)
    assert searched_collection.properties is not None
    assert len(searched_collection.properties) == 0
    assert searched_collection.properties_table is not None


def test_search_no_end_date(region_100ha):
    """ Test MaskedCollection.search() with ``end_date=None`` searches for a day from ``start_date`. """
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    searched_collection = gd_collection.search('2022-01-03', None, region_100ha)

    start_date = datetime.strptime('2022-01-03', '%Y-%m-%d')
    end_date = datetime.strptime('2022-01-04', '%Y-%m-%d')
    properties = searched_collection.properties
    im_dates = np.array(
        [datetime.utcfromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    assert len(properties) > 0
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)


def test_search_date_error(region_100ha):
    """ Test MaskedCollection.search() raises an error when end date is on or before start date. """
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    with pytest.raises(ValueError):
        _ = gd_collection.search('2022-01-02', '2022-01-01', region_100ha)


def test_search_mult_kwargs(region_100ha):
    """
    When a search filtered collection is searched again, test that masks change with different cloud/shadow kwargs.
    """
    start_date = '2022-01-01'
    end_date = '2022-01-10'
    gd_collection = MaskedCollection.from_name('COPERNICUS/S2_SR')

    def get_cloudless_portion(properties: Dict) -> List[float]:
        return [prop_dict['CLOUDLESS_PORTION'] for prop_dict in properties.values()]

    filt_collection = gd_collection.search(start_date, end_date, region_100ha, prob=80, fill_portion=0)
    filt_coll_prob80 = filt_collection.search(start_date, end_date, region_100ha, prob=80, fill_portion=0)
    filt_coll_prob40 = filt_collection.search(start_date, end_date, region_100ha, prob=40, fill_portion=0)

    cp_ref = get_cloudless_portion(filt_collection.properties)
    cp_prob80 = get_cloudless_portion(filt_coll_prob80.properties)
    cp_prob40 = get_cloudless_portion(filt_coll_prob40.properties)

    assert cp_ref == pytest.approx(cp_prob80, abs=1e-3)
    assert cp_ref != pytest.approx(cp_prob40, abs=1e-1)


def test_search_custom_filter(region_25ha):
    """
    Test that a CLOUDLESS_PORTION custom filter gives the same search results as the equivalent cloudless_portion
    kwarg specification.
    """
    start_date = '2022-01-01'
    end_date = '2022-02-01'
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2')
    kwarg_coll = gd_collection.search(start_date, end_date, region_25ha, cloudless_portion=90)
    cust_filt_coll = gd_collection.search(start_date, end_date, region_25ha, custom_filter='CLOUDLESS_PORTION>=90')
    assert (kwarg_coll.properties is not None) and (len(kwarg_coll.properties) > 0)
    assert kwarg_coll.properties == cust_filt_coll.properties


def test_search_add_props(region_25ha):
    """
    Test that specified add_props are added to the search results.
    """
    add_props = ['CLOUD_COVER', 'GEOMETRIC_RMSE_VERIFY']
    gd_collection = MaskedCollection.from_name('LANDSAT/LC09/C02/T1_L2', add_props=add_props)
    searched_coll = gd_collection.search('2022-01-01', '2022-02-01', region_25ha)
    assert all([add_prop in searched_coll.schema.keys() for add_prop in add_props])
    assert all([add_prop in list(searched_coll.properties.values())[0].keys() for add_prop in add_props])
    assert searched_coll.properties_table is not None
    assert searched_coll.schema_table is not None


@pytest.mark.parametrize(
    'name, start_date, end_date, region', [
        ('LANDSAT/LC09/C02/T1_L2', '2022-01-01', '2022-02-01', 'region_100ha'),
        ('COPERNICUS/S2_HARMONIZED', '2022-01-01', '2022-01-15', 'region_100ha'),
        ('LARSE/GEDI/GEDI02_A_002_MONTHLY', '2021-08-01', '2021-09-01', 'region_100ha')
    ]
)
def test_search_no_fill_or_cloudless_portion(
    name: str, start_date: str, end_date: str, region: str, request: pytest.FixtureRequest
):
    """
    Test MaskedCollection.search() without fill / cloudless portion filters gives valid results for different
    collections.
    """
    region: Dict = request.getfixturevalue(region)
    gd_collection = MaskedCollection.from_name(name)
    searched_collection = gd_collection.search(start_date, end_date, region)

    properties = searched_collection.properties
    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')
    im_dates = np.array(
        [datetime.utcfromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
    )
    # test FILL_PORTION and CLOUDLESS_PORTION are not in properties
    prop_keys = list(properties.values())[0].keys()
    assert 'FILL_PORTION' not in prop_keys
    assert 'CLOUDLESS_PORTION' not in prop_keys
    # test search result image dates lie between `start_date` and `end_date`
    assert np.all(im_dates >= start_date) and np.all(im_dates < end_date)
    assert set(searched_collection.schema.keys()) >= set(list(properties.values())[0].keys())
    # test search result image dates are sorted
    assert np.all(sorted(im_dates) == im_dates)


@pytest.mark.parametrize(
    'image_list, method, region, date', [
        ('s2_sr_image_list', CompositeMethod.q_mosaic, 'region_10000ha', None),
        ('s2_sr_image_list', CompositeMethod.q_mosaic, None, '2021-10-01'),
        ('gedi_image_list', CompositeMethod.mosaic, 'region_10000ha', None),
        ('gedi_image_list', CompositeMethod.mosaic, None, '2020-09-01'),
        ('l8_9_image_list', CompositeMethod.medoid, 'region_10000ha', None),
        ('l8_9_image_list', CompositeMethod.medoid, None, '2021-10-01'),
    ]
)
def test_composite_region_date_ordering(image_list, method, region, date, request):
    """
    In MaskedCollection.composite(), test the component images are ordered correctly, according to `region`/`date`
    parameters.
    """
    image_list: List = request.getfixturevalue(image_list)
    region: Dict = request.getfixturevalue(region) if region else None
    gd_collection = MaskedCollection.from_list(image_list)
    ee_collection = gd_collection._prepare_for_composite(method=method, date=date, region=region)
    properties = gd_collection._get_properties(ee_collection)
    assert len(properties) == len(image_list)
    if region:
        # test images are ordered by CLOUDLESS/FILL_PORTION
        schema_keys = list(gd_collection.schema.keys())
        # CLOUDLESS_PORTION is not in MaskedCollection.schema for generic images, so use FILL_PORTION instead
        portion_key = 'CLOUDLESS_PORTION' if 'CLOUDLESS_PORTION' in schema_keys else 'FILL_PORTION'
        im_portions = [im_props[portion_key] for im_props in properties.values()]
        assert sorted(im_portions) == im_portions
    elif date:
        # test images are ordered by time difference with `date`
        im_dates = np.array(
            [datetime.utcfromtimestamp(im_props['system:time_start'] / 1000) for im_props in properties.values()]
        )
        comp_date = datetime.strptime(date, '%Y-%m-%d')
        im_date_diff = np.abs(comp_date - im_dates)
        assert all(sorted(im_date_diff, reverse=True) == im_date_diff)


@pytest.mark.parametrize(
    'image_list, method, mask', [
        ('s2_sr_image_list', CompositeMethod.q_mosaic, True), ('s2_sr_image_list', CompositeMethod.mosaic, False),
        ('l8_9_image_list', CompositeMethod.medoid, True), ('l8_9_image_list', CompositeMethod.median, False),
        ('s2_sr_image_list', CompositeMethod.medoid, True), ('s2_sr_image_list', CompositeMethod.medoid, False),
        ('l8_9_image_list', CompositeMethod.medoid, False),
    ]
)
def test_composite_mask(image_list, method, mask, region_100ha, request):
    """
    Test the MaskedImage.composite() `mask` parameter results in a masked composite image, comprised of masked
    component images.
    """

    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    proj = get_projection(gd_collection.ee_collection.first(), min_scale=True)
    ee_collection = gd_collection._prepare_for_composite(method=method, mask=mask)
    properties = gd_collection._get_properties(ee_collection)
    assert len(properties) == len(image_list)

    def count_masked_pixels(ee_image: ee.Image, count_list: ee.List):
        """ Return the pixel count (area) of the EE mask / valid image area.  """
        ee_mask = ee_image.select('SR_B.*|B.*|rh.*').mask().reduce(ee.Reducer.allNonZero()).rename('EE_MASK')
        count = ee_mask.reduceRegion(reducer='sum', crs=proj.crs(), scale=proj.nominalScale(), geometry=region_100ha)
        return ee.List(count_list).add(count.get('EE_MASK'))

    # get the mask pixel counts for the component images
    component_mask_counts = ee_collection.iterate(count_masked_pixels, ee.List([])).getInfo()
    # get the mask pixel count for the composite image
    composite_im = gd_collection.composite(method=method, mask=mask)
    composite_mask_count = count_masked_pixels(composite_im.ee_image, []).get(0).getInfo()
    if mask:
        # test the composite mask (valid area) is smaller than combined area of the component image masks
        assert composite_mask_count <= np.sum(component_mask_counts)
        # test the composite mask is larger than the smallest of the component image masks
        assert composite_mask_count >= np.min(component_mask_counts)
    else:
        # test that the component mask areas are ~ equal to composite mask area (i.e. the image area)
        assert np.array(component_mask_counts) == pytest.approx(composite_mask_count, rel=.01)


@pytest.mark.parametrize(
    'image_list, resampling, std_scale', [
        ('s2_sr_image_list', ResamplingMethod.bilinear, 60), ('s2_sr_image_list', ResamplingMethod.bicubic, 60),
        ('s2_sr_image_list', ResamplingMethod.average, 120), ('l8_9_image_list', ResamplingMethod.bilinear, 30),
        ('l8_9_image_list', ResamplingMethod.bicubic, 30), ('l8_9_image_list', ResamplingMethod.average, 120),
    ]
)
def test_composite_resampling(
    image_list: str, resampling: ResamplingMethod, std_scale: float, region_10000ha: Dict,
    request: pytest.FixtureRequest
):
    """ Test that resampling smooths the composite image. """

    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    comp_im_before = gd_collection.composite(method=CompositeMethod.mosaic, mask=False)
    comp_im_after = gd_collection.composite(method=CompositeMethod.mosaic, resampling=resampling, mask=False)

    # test the resampled composite is smoother than the default composite
    std_before = get_image_std(comp_im_before.ee_image, region_10000ha, std_scale)
    std_after = get_image_std(comp_im_after.ee_image, region_10000ha, std_scale)
    assert std_before > std_after


def test_composite_s2_cloud_mask_params(s2_sr_image_list, region_10000ha):
    """
    Test cloud/shadow mask **kwargs are passed from MaskedCollection.composite() through to
    Sentinel2ClImage._aux_image().
    """
    gd_collection = MaskedCollection.from_list(s2_sr_image_list)
    comp_im_prob80 = gd_collection.composite(prob=80)
    comp_im_prob80._set_region_stats(region_10000ha, scale=gd_collection.stats_scale)
    comp_im_prob40 = gd_collection.composite(prob=40)
    comp_im_prob40._set_region_stats(region_10000ha, scale=gd_collection.stats_scale)
    prob80_portion = comp_im_prob80.properties['FILL_PORTION']
    prob40_portion = comp_im_prob40.properties['FILL_PORTION']
    assert prob80_portion > prob40_portion


def test_composite_landsat_cloud_mask_params(l8_9_image_list, region_10000ha):
    """
    Test cloud/shadow mask **kwargs are passed from MaskedCollection.composite() through to
    LandsatImage._aux_image().
    """
    gd_collection = MaskedCollection.from_list(l8_9_image_list)
    comp_im_wshadows = gd_collection.composite(mask_shadows=False)
    comp_im_wshadows._set_region_stats(region_10000ha, scale=gd_collection.stats_scale)
    comp_im_woshadows = gd_collection.composite(mask_shadows=True)
    comp_im_woshadows._set_region_stats(region_10000ha, scale=gd_collection.stats_scale)
    with_shadows_portion = comp_im_wshadows.properties['FILL_PORTION']
    without_shadows_portion = comp_im_woshadows.properties['FILL_PORTION']
    assert with_shadows_portion > without_shadows_portion


@pytest.mark.parametrize(
    'image_list, method, mask, region, date, cloud_kwargs', [
        ('s2_sr_image_list', CompositeMethod.q_mosaic, True, 'region_100ha', None, {}),
        ('s2_sr_image_list', CompositeMethod.mosaic, True, None, '2021-10-01', {}),
        ('s2_sr_image_list', CompositeMethod.medoid, False, None, None, {}),
        (
            's2_sr_image_list', CompositeMethod.median, True, None, None,
            dict(
                mask_method='qa', mask_cirrus=False, mask_shadows=False, prob=60, dark=0.2, shadow_dist=500, buffer=500,
                cdi_thresh=None, max_cloud_dist=2000
            )
        ),
        ('l8_9_image_list', CompositeMethod.q_mosaic, True, 'region_100ha', None, {}),
        ('l8_9_image_list', CompositeMethod.mosaic, False, None, '2022-03-01', {}),
        ('l8_9_image_list', CompositeMethod.medoid, True, None, None, dict(mask_cirrus=False, mask_shadows=False)),
        ('l8_9_image_list', CompositeMethod.median, True, None, None, {}),
        ('l4_5_image_list', CompositeMethod.q_mosaic, False, 'region_100ha', None, {}),
        (
            'l4_5_image_list', CompositeMethod.mosaic, True, None, '1988-01-01',
            dict(mask_cirrus=False, mask_shadows=False)
        ),
        ('l4_5_image_list', CompositeMethod.medoid, True, None, None, {}),
        ('l4_5_image_list', CompositeMethod.median, True, None, None, {}),
        ('gedi_image_list', CompositeMethod.mosaic, True, 'region_100ha', None, {}),
        ('gedi_image_list', CompositeMethod.mosaic, True, None, '2020-09-01', {}),
        ('gedi_image_list', CompositeMethod.medoid, True, None, None, {}),
    ]
)  # yapf: disable
def test_composite(image_list, method, mask, region, date, cloud_kwargs, request):
    """ Test MaskedCollection.composite() runs successfully with a variety of `method` and other parameters. """
    image_list: List = request.getfixturevalue(image_list)
    region: Dict = request.getfixturevalue(region) if region else None
    gd_collection = MaskedCollection.from_list(image_list)
    comp_im = gd_collection.composite(method=method, mask=mask, region=region, date=date, **cloud_kwargs)
    assert comp_im._ee_info is not None and len(comp_im._ee_info) > 0
    assert 'INPUT_IMAGES' in comp_im.properties


def test_composite_errors(gedi_image_list, region_100ha):
    """ Test MaskedCollection.composite() error conditions. """
    gedi_collection = MaskedCollection.from_list(gedi_image_list)
    with pytest.raises(ValueError):
        # q-mosaic is only supported on cloud/shadow maskable images
        gedi_collection.composite(method=CompositeMethod.q_mosaic)
    with pytest.raises(ValueError):
        # unknown method
        gedi_collection.composite(method='unknown')
    with pytest.raises(ValueError):
        # date format error
        gedi_collection.composite(method=CompositeMethod.mosaic, date='not a date')
    with pytest.raises(ValueError):
        # composite of empty collection
        empty_collection = gedi_collection.search('2000-01-01', '2000-01-02', region_100ha, 100)
        empty_collection.composite(method=CompositeMethod.mosaic)


@pytest.mark.parametrize('image_list', ['s2_sr_image_list', 'l8_9_image_list'])
def test_composite_date(image_list: str, request: pytest.FixtureRequest):
    """ Test the composite date is the same as the first input image date. """

    image_list: List = request.getfixturevalue(image_list)
    gd_collection = MaskedCollection.from_list(image_list)
    # assumes the image_list's are in date order
    first_date = datetime.utcfromtimestamp(
        gd_collection.ee_collection.first().get('system:time_start').getInfo() / 1000
    ) # yapf: disable
    comp_im = gd_collection.composite()
    assert comp_im.date == first_date


def test_composite_mult_kwargs(region_100ha):
    """
    When a search filtered collection is composited, test that masks change with different cloud/shadow kwargs i.e.
    test that image *_MASK bands are overwritten in the encapsulated collection.
    """
    gd_collection = MaskedCollection.from_name('COPERNICUS/S2_SR')
    filt_collection = gd_collection.search('2022-01-01', '2022-01-10', region_100ha)

    comp_im_prob80 = filt_collection.composite(prob=80)
    comp_im_prob80._set_region_stats(region_100ha, scale=filt_collection.stats_scale)
    comp_im_prob40 = filt_collection.composite(prob=40)
    comp_im_prob40._set_region_stats(region_100ha, scale=filt_collection.stats_scale)

    cp_prob80 = comp_im_prob80.properties['FILL_PORTION']
    cp_prob40 = comp_im_prob40.properties['FILL_PORTION']

    assert cp_prob80 != pytest.approx(cp_prob40, abs=1e-1)


@pytest.mark.parametrize(
    'name', ['FAO/WAPOR/2/L1_RET_E', 'MODIS/006/MCD43A4']
)
def test_unbounded_search_no_region(name):
    """
    Test searching an unbounded collection without a region raises an exception.
    """
    start_date = '2022-01-01'
    end_date = '2022-01-02'
    gd_collection = MaskedCollection.from_name(name)
    gd_collection = gd_collection.search(start_date, end_date, fill_portion=0)
    with pytest.raises(ValueError) as ex:
        _ = gd_collection.properties
    assert 'unbounded' in str(ex) and 'region' in str(ex)


def test_unknown_collection_error(region_25ha):
    """
    Test searching a non-existent collection gives an error.
    """
    start_date = '2022-01-01'
    end_date = '2022-01-02'
    gd_collection = MaskedCollection.from_name('Unknown')
    with pytest.raises(ee.EEException) as ex:
        gd_collection = gd_collection.search(start_date, end_date, region_25ha)
        _ = gd_collection.properties
    assert 'not found' in str(ex)




"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
import re
from collections import OrderedDict
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Union

import ee
import tabulate
import textwrap as wrap
from geedim import schema
from geedim.medoid import medoid
from geedim.download import BaseImage
from geedim.enums import ResamplingMethod, CompositeMethod
from geedim.errors import UnfilteredError, InputImageError
from geedim.mask import MaskedImage, class_from_id
from geedim.stac import StacCatalog, StacItem
from geedim.utils import split_id, resample
from tabulate import TableFormat, Line, DataRow

logger = logging.getLogger(__name__)
tabulate.MIN_PADDING = 0

##
# tabulate format for collection properties
_table_fmt = TableFormat(
    lineabove=Line("", "-", " ", ""),
    linebelowheader=Line("", "-", " ", ""),
    linebetweenrows=None,
    linebelow=Line("", "-", " ", ""),
    headerrow=DataRow("", " ", ""),
    datarow=DataRow("", " ", ""),
    padding=0,
    with_header_hide=["lineabove", "linebelow"]
)  # yapf: disable


def compatible_collections(names: List[str]) -> bool:
    """
    Test if all the collections in a list are spectrally compatible i.e. images from these collections can be
    composited together.
    """
    names = list(set(names))  # reduce to unique values
    start_name = names[0]
    name_matches = [True]
    landsat_regex = re.compile('(LANDSAT/\w{2})(\d{2})(/.*)')
    start_landsat_match = landsat_regex.search(start_name)
    for name in names[1:]:
        name_match = False
        if start_name == name:
            name_match = True
        elif start_landsat_match:
            landsat_regex = re.compile(f'{start_landsat_match.groups()[0]}\d\d{start_landsat_match.groups()[-1]}')
            if landsat_regex.search(name):
                name_match = True
        name_matches.append(name_match)
    return all(name_matches)


def parse_date(date: Union[datetime, str], var_name=None) -> datetime:
    """ Convert a string to a datetime, raising an exception if it is in the wrong format. """
    var_name = var_name or 'date'
    if isinstance(date, str):
        try:
            date = datetime.strptime(date, '%Y-%m-%d')
        except ValueError:
            raise ValueError(f'{var_name} should be a datetime instance or a string with format: "%Y-%m-%d"')
    return date


def abbreviate(name: str) -> str:
    """ Return an acronym for a string in camel or snake case. """
    name = name.strip()
    if len(name) <= 5:
        return name
    abbrev = ''
    prev = '_'
    for curr in name:
        if curr.isdigit():
            abbrev += curr
        elif (prev == '_' and curr.isalnum()) or (prev.islower() and curr.isupper()):
            abbrev += curr.upper()
        prev = curr
    return abbrev if len(abbrev) >= 2 else name


class MaskedCollection:
    _sort_methods = [CompositeMethod.mosaic, CompositeMethod.q_mosaic, CompositeMethod.medoid]

    def __init__(self, ee_collection: ee.ImageCollection, add_props: List[str] = None):
        """
        A class for describing, searching and compositing an Earth Engine image collection, with support for
        cloud/shadow masking.

        Parameters
        ----------
        ee_collection : ee.ImageCollection
            Earth Engine image collection to encapsulate.
        add_props: list of str, optional
            Additional Earth Engine image properties to include in :attr:`properties`.
        """
        if not isinstance(ee_collection, ee.ImageCollection):
            raise TypeError(f'`ee_collection` must be an instance of ee.ImageCollection')
        self._name = None
        self._properties = None
        self._schema = None
        self._filtered = False
        self._ee_collection = ee_collection
        self._add_props = list(add_props) if add_props else None
        self._image_type = None
        self._stac = None
        self._stats_scale = None

    @classmethod
    def from_name(cls, name: str, add_props: List[str] = None) -> 'MaskedCollection':
        """
        Create a MaskedCollection instance from an Earth Engine image collection name.

        Parameters
        ----------
        name: str
            Name of the Earth Engine image collection to create.
        add_props: list of str, optional
            Additional Earth Engine image properties to include in :attr:`properties`.

        Returns
        -------
        MaskedCollection
            A MaskedCollection instance.
        """
        # this is separate from __init__ for consistency with MaskedImage.from_id()
        if (name == 'COPERNICUS/S2') or (name == 'COPERNICUS/S2_SR'):
            # Recent images in S2_SR do not always have matching images in S2_CLOUD_PROBABILITY (which is needed for
            # 'cloud_prob' cloud masking), so this is a special case to return a filtered S2/S2_SR collection that has
            # matching images in S2_CLOUD_PROBABILITY.
            cloud_prob_coll = ee.ImageCollection('COPERNICUS/S2_CLOUD_PROBABILITY')
            s2_coll = ee.ImageCollection(name)
            filt = ee.Filter.equals(leftField='system:index', rightField='system:index')
            ee_collection = ee.ImageCollection(ee.Join.simple().apply(s2_coll, cloud_prob_coll, filt))
        else:
            ee_collection = ee.ImageCollection(name)
        gd_collection = cls(ee_collection, add_props=add_props)
        gd_collection._name = name
        return gd_collection

    @classmethod
    def from_list(
        cls, image_list: List[Union[str, MaskedImage, ee.Image]], add_props: List[str] = None
    ) -> 'MaskedCollection':
        """
        Create a MaskedCollection instance from a list of Earth Engine image ID strings, ``ee.Image`` instances and/or
        :class:`~geedim.mask.MaskedImage` instances.  The list may include composite images, as created with
        :meth:`composite`.

        Images from spectrally compatible Landsat collections can be combined i.e. Landsat-4 with Landsat-5,
        and Landsat-8 with Landsat-9.  Otherwise, images should all belong to the same collection.

        Any ``ee.Image`` instances in the list (including those encapsulated by :class:`~geedim.mask.MaskedImage`)
        must have ``system:id`` and ``system:time_start`` properties.  This is always the case for images
        obtained from the Earth Engine catalog, and images returned from :meth:`composite`.  Any other user-created
        images passed to :meth:`from_list` should have these properties set.

        Parameters
        ----------
        image_list : list
            List of images to include in the collection (must all be from the same, or compatible Earth Engine
            collections). List items can be ID strings, instances of :class:`~geedim.mask.MaskedImage`, or instances of
            ``ee.Image``.
        add_props: list of str, optional
            Additional Earth Engine image properties to include in :attr:`properties`.

        Returns
        -------
        MaskedCollection
            A MaskedCollection instance.
        """
        # build lists of EE images and IDs from image_list
        if len(image_list) == 0:
            raise ValueError('`image_list` is empty.')

        im_dict_list = []
        for image_obj in image_list:
            if isinstance(image_obj, str):
                im_dict_list.append(dict(ee_image=ee.Image(image_obj), id=image_obj, has_date=True))
            elif isinstance(image_obj, ee.Image):
                ee_info = image_obj.getInfo()
                ee_id = ee_info['id'] if 'id' in ee_info else None
                has_date = ('properties' in ee_info) and ('system:time_start' in ee_info['properties'])
                im_dict_list.append(dict(ee_image=ee.Image(image_obj), id=ee_id, has_date=has_date))
            elif isinstance(image_obj, BaseImage):
                im_dict_list.append(
                    dict(ee_image=image_obj.ee_image, id=image_obj.id, has_date=image_obj.date is not None)
                )
            else:
                raise TypeError(f'Unsupported image object type: {type(image_obj)}')

        # check all images have IDs and capture dates
        if any([(im_dict['id'] is None) or (not im_dict['has_date']) for im_dict in im_dict_list]):
            raise InputImageError('Image(s) must have "id" and "system:time_start" properties.')

        # check the images all come from the same or compatible collections
        ee_coll_names = [split_id(im_dict['id'])[0] for im_dict in im_dict_list]
        if not compatible_collections(ee_coll_names):
            raise InputImageError('All images must belong to the same, or spectrally compatible, collections.')

        # create the collection object, using the name of the first collection in ee_coll_names.
        gd_collection = cls.from_name(ee_coll_names[0], add_props=add_props)
        gd_collection._ee_collection = ee.ImageCollection(ee.List([im_dict['ee_image'] for im_dict in im_dict_list]))
        gd_collection._filtered = True
        return gd_collection

    @property
    def stac(self) -> Union[StacItem, None]:
        """ STAC info, if any.  """
        if not self._stac and (self.name in StacCatalog().url_dict):
            self._stac = StacCatalog().get_item(self.name)
        return self._stac

    @property
    def stats_scale(self) -> Union[float, None]:
        """ Scale to use for re-projections when finding region statistics. """
        if not self.stac:
            return None
        if not self._stats_scale:
            gsds = [float(band_dict['gsd']) for band_dict in self.stac.band_props.values()]
            max_gsd = max(gsds)
            min_gsd = min(gsds)
            self._stats_scale = min_gsd if (max_gsd > 10 * min_gsd) and (min_gsd > 0) else max_gsd
        return self._stats_scale

    @property
    def ee_collection(self) -> ee.ImageCollection:
        """ Earth Engine image collection. """
        return self._ee_collection

    @property
    def name(self) -> str:
        """ Name of the encapsulated Earth Engine image collection. """
        if not self._name:
            ee_info = self._ee_collection.first().getInfo()
            self._name = split_id(ee_info['id'])[0] if ee_info and 'id' in ee_info else 'None'
        return self._name

    @property
    def image_type(self) -> type:
        """ :class:`~geedim.mask.MaskedImage` class or sub-class corresponding to images in :attr:`ee_collection`. """
        if not self._image_type:
            self._image_type = class_from_id(self.name)
        return self._image_type

    @property
    def properties(self) -> Dict[str, Dict]:
        """
        A dictionary of properties for each image in the collection. Dictionary keys are the image IDs, and values
        are a dictionaries of image properties.
        """
        if not self._filtered:
            raise UnfilteredError(
                '`properties` can only be retrieved for collections returned by `search()` and `from_list()`'
            )
        if not self._properties:
            self._properties = self._get_properties(self._ee_collection)
        return self._properties

    @property
    def properties_table(self) -> str:
        """ :attr:`properties` formatted as a printable table string. """
        return self._get_properties_table(self.properties)

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        A dictionary of abbreviations and descriptions for :attr:`properties`. Keys are the EE property names,
        and values are dictionaries of abbreviations and descriptions.
        """
        if not self._schema:
            if self.name in schema.collection_schema:
                self._schema = schema.collection_schema[self.name]['prop_schema'].copy()
            else:
                self._schema = schema.default_prop_schema.copy()

            # append any additional properties to the schema
            if self._add_props:
                for add_prop in self._add_props:
                    # get a description from STAC where possible
                    description = (
                        self.stac.descriptions[add_prop] if self.stac and add_prop in self.stac.descriptions else ''
                    )
                    # remove newlines from description and crop to the first sentence
                    description = ' '.join(description.strip().splitlines()).split('. ')[0]
                    self._schema[add_prop] = dict(abbrev=abbreviate(add_prop), description=description)
        return self._schema

    @property
    def schema_table(self) -> str:
        """ :attr:`schema` formatted as a printable table string. """
        table_list = []
        for prop_name, prop_dict in self.schema.items():
            description = '\n'.join(wrap.wrap(prop_dict['description'], 50))
            table_list.append(dict(abbrev=prop_dict['abbrev'], name=prop_name, description=description))
        headers = {key: key.upper() for key in table_list[0].keys()}
        return tabulate.tabulate(table_list, headers=headers, floatfmt='.2f', tablefmt='simple')

    @property
    def refl_bands(self) -> Union[List[str], None]:
        """ List of spectral / reflectance band names, if any. """
        if not self.stac:
            return None
        return [bname for bname, bdict in self.stac.band_props.items() if 'center_wavelength' in bdict]

    def _get_properties(self, ee_collection: ee.ImageCollection, schema=None) -> Dict:
        """ Retrieve properties of images in a given Earth Engine image collection. """
        if not schema:
            schema = self.schema

        # the properties to retrieve
        prop_key_list = ee.List(list(schema.keys()))

        def aggregrate_props(ee_image, coll_list):
            im_dict = ee_image.toDictionary(prop_key_list)
            return ee.List(coll_list).add(im_dict)

        # retrieve list of dicts of properties of images in ee_collection
        props_list = []
        try:
            props_list = ee.List(ee_collection.iterate(aggregrate_props, ee.List([]))).getInfo()
        except ee.EEException as ex:
            if 'geometry' in str(ex) and 'unbounded' in str(ex):
                raise ValueError(
                    'This collection is unbounded and needs a `region` to be specified to find FILL_PORTION / '
                    'CLOUDLESS_PORTION.'
                )
            else:
                raise
        # add image properties to the return dict in the same order as the underlying collection
        props_dict = OrderedDict()
        for prop_dict in props_list:
            props_dict[prop_dict['system:id']] = prop_dict
        return props_dict

    def _get_properties_table(self, properties: Dict, schema: Dict = None) -> str:
        """
        Format the given properties into a table.  Orders properties (columns) according to :attr:`schema` and
        replaces long form property names with abbreviations.
        """
        if not schema:
            schema = self.schema

        abbrev_props = []
        for im_id, im_prop_dict in properties.items():
            abbrev_dict = OrderedDict()
            for prop_name, key_dict in schema.items():
                if prop_name in im_prop_dict:
                    if prop_name == 'system:time_start':  # convert timestamp to date string
                        dt = datetime.utcfromtimestamp(im_prop_dict[prop_name] / 1000)
                        abbrev_dict[key_dict['abbrev']] = datetime.strftime(dt, '%Y-%m-%d %H:%M')
                    else:
                        abbrev_dict[key_dict['abbrev']] = im_prop_dict[prop_name]
            abbrev_props.append(abbrev_dict)
        return tabulate.tabulate(abbrev_props, headers='keys', floatfmt='.2f', tablefmt=_table_fmt)

    def _prepare_for_composite(
        self, method: CompositeMethod, mask: bool = True, resampling: Union[ResamplingMethod, str] = None,
        date: str = None, region: Dict = None, **kwargs
    ) -> ee.ImageCollection:
        """
        Prepare the Earth Engine collection for compositing. See :meth:`~MaskedCollection.composite` for
        parameter descriptions.
        """
        date = parse_date(date, 'date')

        if not self._filtered:
            raise UnfilteredError(
                'Composites can only be created from collections returned by `search()` and `from_list()`'
            )

        method = CompositeMethod(method)
        resampling = ResamplingMethod(resampling) if resampling else BaseImage._default_resampling
        if (method == CompositeMethod.q_mosaic) and (self.image_type == MaskedImage):
            cloud_coll_str = ", ".join(schema.cloud_coll_names)
            raise ValueError(
                f'The `q-mosaic` method is not supported for this collection: {self.name}.  It is supported for the '
                f'{cloud_coll_str} collections only.'
            )

        def prepare_image(ee_image: ee.Image):
            """ Prepare an Earth Engine image for use in compositing. """
            if date and (method in self._sort_methods):
                date_dist = ee.Number(ee_image.get('system:time_start')).subtract(ee.Date(date).millis()).abs()
                ee_image = ee_image.set('DATE_DIST', date_dist)

            gd_image = self.image_type(ee_image, **kwargs)
            if region and (method in self._sort_methods):
                gd_image._set_region_stats(region=region, scale=self.stats_scale)
            if mask:
                gd_image.mask_clouds()
            return resample(gd_image.ee_image, resampling)

        ee_collection = self._ee_collection.map(prepare_image)

        if method in self._sort_methods:
            if date:
                ee_collection = ee_collection.sort('DATE_DIST', opt_ascending=False)
            elif region:
                sort_key = 'CLOUDLESS_PORTION' if 'CLOUDLESS_PORTION' in self.schema.keys() else 'FILL_PORTION'
                ee_collection = ee_collection.sort(sort_key)
            else:
                ee_collection = ee_collection.sort('system:time_start')
        else:
            if date:
                logger.warning('`date` is valid for `mosaic`, `q_mosaic` and `medoid` methods only.')
            elif region:
                logger.warning('`region` is valid for `mosaic`, `q_mosaic` and `medoid` methods only.')

        return ee_collection

    def search(
        self, start_date: Union[datetime, str] = None, end_date: Union[datetime, str] = None, region: Dict = None,
        fill_portion: float = None, cloudless_portion: float = None, custom_filter: str = None, **kwargs
    ) -> 'MaskedCollection':
        """
        Search for images based on date, region, filled/cloudless portion, and custom criteria.

        Filled and cloudless portions are only calculated and included in collection properties when one or both of
        ``fill_portion`` / ``cloudless_portion`` are specified.

        Search speed can be increased by specifying ``custom_filter``, and or by omitting ``fill_portion`` /
        ``cloudless_portion``.

        Parameters
        ----------
        start_date : datetime, str
            Start date (UTC).  In '%Y-%m-%d' format if a string.
        end_date : datetime, str, optional
            End date (UTC).  In '%Y-%m-%d' format if a string.  If None, ``end_date`` is set to a day after
            ``start_date``.
        region : dict, ee.Geometry
            Polygon in WGS84 specifying a region that images should intersect.
        fill_portion: float, optional
            Lower limit on the portion of region that contains filled/valid image pixels (%).
        cloudless_portion: float, optional
            Lower limit on the portion of filled pixels that are cloud/shadow free (%).
        custom_filter: str, optional
            Custom image property filter expression e.g. "property > value".  See the `EE docs
            <https://developers.google.com/earth-engine/apidocs/ee-filter-expression>`_.
        **kwargs
            Optional cloud/shadow masking parameters - see :meth:`geedim.mask.MaskedImage.__init__` for details.

        Returns
        -------
        MaskedCollection
            Filtered MaskedCollection instance containing the search result image(s).
        """
        if not start_date and not region:
            raise ValueError('At least one of `start_date` or `region` should be specified')

        if not start_date or not region:
            logger.warning('Specifying `start_date` and `region` will improve the search speed.')

        def set_region_stats(ee_image: ee.Image):
            """ Find filled and cloud/shadow free portions inside the search region for a given image.  """
            gd_image = self.image_type(ee_image, **kwargs)
            gd_image._set_region_stats(region, scale=self.stats_scale)
            return gd_image.ee_image

        # filter the image collection, finding cloud/shadow masks and region stats
        ee_collection = self._ee_collection
        if start_date:
            start_date = parse_date(start_date, 'start_date')
            if end_date is None:
                # set end_date a day later than start_date
                end_date = start_date + timedelta(days=1)
            else:
                end_date = parse_date(end_date, 'end_date')
            if end_date <= start_date:
                raise ValueError('`end_date` must be at least a day later than `start_date`')
            ee_collection = ee_collection.filterDate(start_date, end_date)

        if region:
            ee_collection = ee_collection.filterBounds(region)

        # when possible filter on custom_filter before calling set_region_stats to reduce computation
        if custom_filter and all([prop_key not in custom_filter for prop_key in ['FILL_PORTION', 'CLOUDLESS_PORTION']]):
            ee_collection = ee_collection.filter(ee.Filter.expression(custom_filter))
            custom_filter = None

        if (fill_portion is not None) or (cloudless_portion is not None) or (custom_filter is not None):
            # set regions stats before filtering on those properties
            ee_collection = ee_collection.map(set_region_stats)
            if fill_portion:
                ee_collection = ee_collection.filter(ee.Filter.gte('FILL_PORTION', fill_portion))

            if cloudless_portion and self.image_type != MaskedImage:
                ee_collection = ee_collection.filter(ee.Filter.gte('CLOUDLESS_PORTION', cloudless_portion))

            # filter on custom_filter that refers to FILL_ or CLOUDLESS_PORTION
            if custom_filter:
                # this expression can include properties from set_region_stats
                ee_collection = ee_collection.filter(ee.Filter.expression(custom_filter))

        ee_collection = ee_collection.sort('system:time_start')

        # return a new MaskedCollection containing the filtered EE collection (the EE collection
        # wrapped by MaskedCollection remains fixed)
        gd_collection = MaskedCollection(ee_collection, add_props=self._add_props)
        gd_collection._name = self._name
        gd_collection._filtered = True
        return gd_collection

    def composite(
        self, method: Union[CompositeMethod, str] = None, mask: bool = True,
        resampling: Union[ResamplingMethod, str] = None, date: Union[datetime, str] = None, region: dict = None,
        **kwargs
    ) -> MaskedImage:
        """
        Create a composite image from the encapsulated image collection.

        Parameters
        ----------
        method: CompositeMethod, str, optional
            Method for finding each composite pixel from the stack of corresponding input image pixels. See
            :class:`~geedim.enums.CompositeMethod` for available options.  By default, `q-mosaic` is used for
            cloud/shadow mask supported collections, `mosaic` otherwise.
        mask: bool, optional
            Whether to apply the cloud/shadow mask; or fill (valid pixel) mask, in the case of images without
            support for cloud/shadow masking.
        resampling: ResamplingMethod, str, optional
            Resampling method to use on collection images prior to compositing.  If None, `near` resampling is used
            (the default).  See :class:`~geedim.enums.ResamplingMethod` for available options.
        date: datetime, str, optional
            Sort collection images by their absolute difference in capture time from this date.  Useful for
            prioritising pixels from images closest to this date.  Valid for the `q-mosaic`, `mosaic` and
            `medoid` ``method`` only.  If None, no time difference sorting is done (the default).
        region: dict, optional
            Sort collection images by the portion of their pixels that are cloudless, and inside this geojson polygon.
            This is useful to prioritise pixels from the least cloudy image(s). Valid for the `q-mosaic` `mosaic`, and
            `medoid` ``method`` only.  If collection has no cloud/shadow mask support, images are sorted by the portion
            of their pixels that are valid, and inside ``region``. If None, no cloudless/valid portion sorting is done
            (the default).  If ``date`` and ``region`` are not specified, collection images are sorted by their capture
            date.
        **kwargs
            Optional cloud/shadow masking parameters - see :meth:`geedim.mask.MaskedImage.__init__` for details.

        Returns
        -------
        MaskedImage
            Composite image.
        """

        if isinstance(date, str):
            try:
                date = datetime.strptime(date, '%Y-%m-%d')
            except ValueError:
                raise ValueError('`date` should be a datetime instance or a string with format: "%Y-%m-%d"')

        if method is None:
            method = CompositeMethod.mosaic if self.image_type == MaskedImage else CompositeMethod.q_mosaic

        # mask, sort & resample the EE collection
        method = CompositeMethod(method)
        ee_collection = self._prepare_for_composite(
            method=method, mask=mask, resampling=resampling, date=date, region=region, **kwargs
        )

        if method == CompositeMethod.q_mosaic:
            comp_image = ee_collection.qualityMosaic('CLOUD_DIST')
        elif method == CompositeMethod.mosaic:
            comp_image = ee_collection.mosaic()
        elif method == CompositeMethod.median:
            comp_image = ee_collection.median()
        elif method == CompositeMethod.medoid:
            # limit medoid to surface reflectance bands
            comp_image = medoid(ee_collection, bands=self.refl_bands)
        elif method == CompositeMethod.mode:
            comp_image = ee_collection.mode()
        elif method == CompositeMethod.mean:
            comp_image = ee_collection.mean()
        else:
            raise ValueError(f'Unsupported composite method: {method}')

        # populate composite image metadata with info on component images
        # TODO: speed this up, e.g. can we use just ID's and times from self.properties?
        props = self._get_properties(ee_collection)
        if len(props) == 0:
            raise ValueError('The collection is empty.')
        props_str = self._get_properties_table(props)
        comp_image = comp_image.set('INPUT_IMAGES', 'TABLE:\n' + props_str)

        # construct an ID for the composite
        dates = [datetime.utcfromtimestamp(item['system:time_start'] / 1000) for item in props.values()]
        start_date = min(dates).strftime('%Y_%m_%d')
        end_date = max(dates).strftime('%Y_%m_%d')

        method_str = method.value.upper()
        if method in self._sort_methods and date:
            method_str += '-' + date.strftime('%Y_%m_%d')

        comp_id = f'{self.name}/{start_date}-{end_date}-{method_str}-COMP'
        comp_image = comp_image.set('system:id', comp_id)  # sets root 'id' property
        comp_image = comp_image.set('system:index', comp_id)  # sets 'properties'->'system:index'

        # set the composite capture time to the capture time of the first input image.
        # note that we must specify the timezone as utc, otherwise .timestamp() assumes local time.
        timestamp = min(dates).replace(tzinfo=timezone.utc).timestamp() * 1000
        comp_image = comp_image.set('system:time_start', timestamp)
        gd_comp_image = self.image_type(comp_image)
        gd_comp_image._id = comp_id  # avoid getInfo() for id property
        return gd_comp_image



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import re

import pytest
from geedim.stac import StacCatalog, StacItem
from geedim.utils import split_id


@pytest.fixture(scope='session')
def stac_catalog() -> StacCatalog:
    """ The StacCatalog instance.  """
    return StacCatalog()


@pytest.fixture(scope='session')
def s2_sr_stac_item(stac_catalog, s2_sr_image_id) -> StacItem:
    """ A StacItem for the Sentinel-2 SR collection.  """
    return stac_catalog.get_item(s2_sr_image_id)


def test_singleton(landsat_ndvi_image_id: str):
    """ Test StacCatalog is a singleton. """
    coll_name, _ = split_id(landsat_ndvi_image_id)
    stac1 = StacCatalog()
    stac2 = StacCatalog()
    _ = stac1.url_dict
    assert len(stac2._url_dict) > 0
    _ = stac1.get_item(coll_name)
    assert coll_name in stac2._cache


def test_traverse_stac(stac_catalog: StacCatalog):
    """ Test _traverse_stac() on the root of the COPERNICUS subtree. """
    url_dict = {}
    url_dict = stac_catalog._traverse_stac(
        'https://storage.googleapis.com/earthengine-stac/catalog/COPERNICUS/catalog.json', url_dict
    )
    assert len(url_dict) > 0
    assert 'COPERNICUS/S2_SR' in url_dict


@pytest.mark.parametrize(
    'image_id', [
        'l4_image_id', 'l5_image_id', 'l7_image_id', 'l8_image_id', 'l9_image_id', 'landsat_ndvi_image_id',
        's2_sr_image_id', 's2_toa_image_id', 's1_sar_image_id', 'modis_nbar_image_id', 'gedi_cth_image_id',
    ]
)
def test_known_get_item(image_id: str, stac_catalog: StacCatalog, request: pytest.FixtureRequest):
    """  Test that stac_catalog contains expected 'items'.  """
    image_id = request.getfixturevalue(image_id)
    coll_name, _ = split_id(image_id)
    assert coll_name in stac_catalog.url_dict
    stac_item = stac_catalog.get_item(coll_name)
    assert stac_item is not None


def test_unknown_get_item(stac_catalog: StacCatalog):
    """  Test that stac_catalog returns None for unknown entries.  """
    assert stac_catalog.get_item_dict('unknown') is None
    assert stac_catalog.get_item('unknown') is None


@pytest.mark.parametrize(
    'image_id', [
        'l4_image_id', 'l5_image_id', 'l7_image_id', 'l8_image_id', 'l9_image_id', 's2_sr_image_id', 's2_toa_image_id',
        's2_sr_hm_image_id', 's2_toa_hm_image_id', 'modis_nbar_image_id',
    ]
)
def test_refl_stac_item(image_id: str, stac_catalog: StacCatalog, request: pytest.FixtureRequest):
    """ Test reflectance collectionStacItem properties are as expected. """
    image_id = request.getfixturevalue(image_id)
    coll_name, _ = split_id(image_id)
    stac_item = stac_catalog.get_item(coll_name)
    assert stac_item is not None
    if coll_name:
        assert stac_item.name == coll_name
    assert len(stac_item.license) > 0
    assert stac_item.band_props is not None
    for key in ['gsd', 'description']:
        has_key = [key in bd for bd in stac_item.band_props.values()]
        assert all(has_key)
    has_center_wavelength = ['center_wavelength' in bd for bd in stac_item.band_props.values()]
    assert sum(has_center_wavelength) >= 7
    for band_dict in stac_item.band_props.values():
        if re.search('^B\d|^SR_B\d|^Nadir_Reflectance_Band\d', band_dict['name']):
            assert 'center_wavelength' in band_dict
            assert 'scale' in band_dict


@pytest.mark.parametrize('image_id', ['l4_image_id', 's2_sr_image_id', 's1_sar_image_id'])
def test_stac_item_descriptions(image_id: str, stac_catalog: StacCatalog, request: pytest.FixtureRequest):
    """ Test StacItem.descriptions. """
    image_id = request.getfixturevalue(image_id)
    coll_name, _ = split_id(image_id)
    stac_item = stac_catalog.get_item(coll_name)
    assert stac_item is not None
    assert stac_item.descriptions is not None
    assert len(stac_item.descriptions) > 0
    assert len(list(stac_item.descriptions.values())[0]) > 0



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
from typing import Dict

import ee
import geedim.schema
from geedim.download import BaseImage
from geedim.enums import CloudMaskMethod
from geedim.utils import split_id, get_projection

logger = logging.getLogger(__name__)

##


class MaskedImage(BaseImage):
    _default_mask = False

    def __init__(self, ee_image: ee.Image, mask: bool = _default_mask, region: dict = None, **kwargs):
        """
        A class for describing, masking and downloading an Earth Engine image.

        Parameters
        ----------
        ee_image: ee.Image
            Earth Engine image to encapsulate.
        mask: bool, optional
            Whether to mask the image.
        region: dict, optional
            A geojson polygon inside of which to find statistics for the image.  These values are stored in the image
            properties. If None, statistics are not found (the default).

        **kwargs
            Cloud/shadow masking parameters - see below:
        mask_cirrus: bool, optional
            Whether to mask cirrus clouds.  Valid for Landsat 8-9 images, and for Sentinel-2 images with
            the `qa` ``mask_method``.
        mask_shadows: bool, optional
            Whether to mask cloud shadows.
        mask_method: CloudMaskMethod, str, optional
            Method used to mask clouds.  Valid for Sentinel-2 images.  See :class:`~geedim.enums.CloudMaskMethod` for
            details.
        prob: float, optional
            Cloud probability threshold (%). Valid for Sentinel-2 images with the `cloud-prob` ``mask-method``.
        dark: float, optional
            NIR threshold [0-1]. NIR values below this threshold are potential cloud shadows.  Valid for Sentinel-2
            images.
        shadow_dist: int, optional
            Maximum distance (m) to look for cloud shadows from cloud edges.  Valid for Sentinel-2 images.
        buffer: int, optional
            Distance (m) to dilate cloud/shadow.  Valid for Sentinel-2 images.
        cdi_thresh: float, optional
            Cloud Displacement Index threshold. Values below this threshold are considered potential clouds.
            If this parameter is not specified (=None), the index is not used.  Valid for Sentinel-2 images.
            See https://developers.google.com/earth-engine/apidocs/ee-algorithms-sentinel2-cdi for details.
        max_cloud_dist: int, optional
            Maximum distance (m) to look for clouds when forming the 'cloud distance' band.  Valid for
            Sentinel-2 images.
        """
        # TODO: consider adding proj_scale parameter here, rather than in _set_region_stats, then it can be re-used in
        #  S2 cloud masking and distance
        BaseImage.__init__(self, ee_image)
        self._add_aux_bands(**kwargs)  # add any mask and cloud distance bands
        if region:
            self._set_region_stats(region)
        if mask:
            self.mask_clouds()

    @staticmethod
    def from_id(image_id: str, **kwargs) -> 'MaskedImage':
        """
        Create a MaskedImage, or sub-class, instance from an Earth Engine image ID.

        Parameters
        ----------
        image_id: str
            ID of the Earth Engine image to encapsulate.
        **kwargs
            Optional keyword arguments to pass to :meth:`__init__`.

        Returns
        -------
        MaskedImage
            A MaskedImage, or sub-class instance.
        """

        cls = class_from_id(image_id)
        ee_image = ee.Image(image_id)
        gd_image = cls(ee_image, **kwargs)
        gd_image._id = image_id  # set the id attribute (avoids a call to getInfo() for .id property)
        return gd_image

    def _aux_image(self, **kwargs) -> ee.Image:
        """
        Retrieve the auxiliary image (MaskedImage provides an image with a FILL_MASK band only).

        Derived classes should override this method and return an image with FILL_MASK and any other auxiliary bands
        they support.
        """
        return self.ee_image.mask().reduce(ee.Reducer.allNonZero()).rename('FILL_MASK')

    def _add_aux_bands(self, **kwargs):
        """
        Add auxiliary bands to the encapsulated image. Existing auxiliary bands are overwritten, unless the image has
        no fixed projection and existing auxiliary bands (i.e. is likely a MaskedCollection composite image).
        """
        aux_image = self._aux_image(**kwargs)
        proj = self.ee_image.select(0).projection()
        has_fixed_scale = proj.nominalScale().toInt64().neq(111319)  # 1 deg in meters
        has_no_aux_bands = ee.Number(self.ee_image.bandNames().indexOf('FILL_MASK').lt(0))
        # overwrite unless it is a composite image with existing aux bands
        overwrite = has_no_aux_bands.Or(has_fixed_scale)
        self.ee_image = ee.Image(
            ee.Algorithms.If(overwrite, self.ee_image.addBands(aux_image, overwrite=True), self.ee_image)
        )

    def _set_region_stats(self, region: Dict = None, scale: float = None):
        """
        Set FILL_PORTION and CLOUDLESS_PORTION on the encapsulated image for the specified region.  Derived classes
        should override this method and set CLOUDLESS_PORTION, and/or other statistics they support.

        Parameters
        ----------
        region : dict, ee.Geometry, optional
            Region inside of which to find statistics.  If not specified, the image footprint is used.
        scale: float, optional
            Re-project to this scale when finding statistics.
        """
        if not region:
            region = self.ee_image.geometry()  # use the image footprint

        proj = get_projection(self.ee_image, min_scale=False)  # get projection of minimum scale band
        # If _proj_scale is set, use that as the scale, otherwise use the proj.nomimalScale().  For non-composite images
        # these should be the same value.  For composite images, there is no `fixed` projection, hence the
        # need for _proj_scale.
        scale = scale or proj.nominalScale()

        # Find the fill portion as the (sum over the region of FILL_MASK) divided by (sum over the region of a constant
        # image (==1)).  We take this approach rather than using a mean reducer, as this does not find the mean over
        # the region, but the mean over the part of the region covered by the image.
        stats_image = ee.Image(
            [self.ee_image.select('FILL_MASK').unmask(), ee.Image(1).rename('REGION_SUM')]
        )  # yapf: disable
        # Note: sometimes proj has no EPSG in crs(), hence use crs=proj and not crs=proj.crs() below
        sums = stats_image.reduceRegion(
            reducer="sum", geometry=region, crs=proj, scale=scale, bestEffort=True, maxPixels=1e6
        )

        fill_portion = (
            ee.Number(sums.get('FILL_MASK')).divide(ee.Number(sums.get('REGION_SUM'))).multiply(100)
        )

        # set the encapsulated image properties
        self.ee_image = self.ee_image.set('FILL_PORTION', fill_portion)
        # set CLOUDLESS_PORTION=100 for the generic case, where cloud/shadow masking is not supported
        self.ee_image = self.ee_image.set('CLOUDLESS_PORTION', 100.)

    def mask_clouds(self):
        """ Apply the cloud/shadow mask if supported, otherwise apply the fill mask. """
        self.ee_image = self.ee_image.updateMask(self.ee_image.select('FILL_MASK'))


class CloudMaskedImage(MaskedImage):
    """ A base class for encapsulating cloud/shadow masked images. """

    def _cloud_dist(self, cloudless_mask: ee.Image = None, max_cloud_dist: float = 5000) -> ee.Image:
        """ Find the cloud/shadow distance in units of 10m. """
        if not cloudless_mask:
            cloudless_mask = self.ee_image.select('CLOUDLESS_MASK')
        proj = get_projection(self.ee_image, min_scale=False)  # use maximum scale projection to save processing time

        # Note that initial *MASK bands before any call to mask_clouds(), are themselves masked, so this cloud/shadow
        # mask excludes (i.e. masks) already masked pixels.  This avoids finding distance to e.g. scanline errors in
        # Landsat-7.
        cloud_shadow_mask = cloudless_mask.Not()
        cloud_pix = ee.Number(max_cloud_dist).divide(proj.nominalScale()).round()  # cloud_dist in pixels

        # Find distance to nearest cloud/shadow (units of 10m).
        cloud_dist = cloud_shadow_mask.fastDistanceTransform(
            neighborhood=cloud_pix, units='pixels', metric='squared_euclidean'
        ).sqrt().multiply(proj.nominalScale().divide(10))

        # Reproject to force calculation at correct scale.
        cloud_dist = cloud_dist.reproject(crs=proj, scale=proj.nominalScale()).rename('CLOUD_DIST')

        # Clip cloud_dist to max_cloud_dist.
        cloud_dist = cloud_dist.where(cloud_dist.gt(ee.Image(max_cloud_dist / 10)), max_cloud_dist / 10)

        # cloud_dist is float64 by default, so convert to Uint16 here to avoid forcing the whole image to float64 on
        # download.
        return cloud_dist.toUint16().rename('CLOUD_DIST')

    def _set_region_stats(self, region: Dict = None, scale: float = None):
        """
        Set FILL_PORTION and CLOUDLESS_PORTION on the encapsulated image for the specified region.

        Parameters
        ----------
        region : dict, ee.Geometry, optional
            Region inside of which to find statistics.  If not specified, the image footprint is used.
        scale: float, optional
            Re-project to this scale when finding statistics.
        """
        if not region:
            region = self.ee_image.geometry()  # use the image footprint

        proj = get_projection(self.ee_image, min_scale=False)  # get projection of minimum scale band
        # If _proj_scale is set, use that as the scale, otherwise use the proj.nomimalScale().  For non-composite images
        # these should be the same value.  For composite images, there is no `fixed` projection, hence the
        # need for _proj_scale.
        scale = scale or proj.nominalScale()

        # Find the fill portion as the (sum over the region of FILL_MASK) divided by (sum over the region of a constant
        # image (==1)).  We take this approach rather than using a mean reducer, as this does not find the mean over
        # the region, but the mean over the part of the region covered by the image.  Then we find cloudless portion
        # as portion of fill that is cloudless.
        stats_image = ee.Image(
            [self.ee_image.select(['FILL_MASK', 'CLOUDLESS_MASK']).unmask(), ee.Image(1).rename('REGION_SUM')]
        )  # yapf: disable
        sums = stats_image.reduceRegion(
            reducer="sum", geometry=region, crs=proj, scale=scale, bestEffort=True, maxPixels=1e6
        ).rename(['FILL_MASK', 'CLOUDLESS_MASK'], ['FILL_PORTION', 'CLOUDLESS_PORTION'])
        fill_portion = (
            ee.Number(sums.get('FILL_PORTION')).divide(ee.Number(sums.get('REGION_SUM'))).multiply(100)
        )
        cloudless_portion = (
            ee.Number(sums.get('CLOUDLESS_PORTION')).divide(ee.Number(sums.get('FILL_PORTION'))).multiply(100)
        )

        # set the encapsulated image properties
        region_stats = ee.Dictionary(dict(FILL_PORTION=fill_portion, CLOUDLESS_PORTION=cloudless_portion))
        self.ee_image = self.ee_image.set(region_stats)

    def _aux_image(self, **kwargs) -> ee.Image:
        """
        Retrieve the auxiliary image containing cloud/shadow masks and cloud distance.  The returned image should
        contain at least FILL_MASK, CLOUDLESS_MASK and CLOUD_DIST bands.

        See :meth:`LandsatImage._aux_image` for an example.
        """
        raise NotImplementedError('This virtual method should be overridden by derived classes.')

    def mask_clouds(self):
        """ Apply the cloud/shadow mask. """
        self.ee_image = self.ee_image.updateMask(self.ee_image.select('CLOUDLESS_MASK'))


class LandsatImage(CloudMaskedImage):
    """
    Class for cloud/shadow masking of Landsat level 2, collection 2 images.

    Supports images from:
    * LANDSAT/LT04/C02/T1_L2
    * LANDSAT/LT05/C02/T1_L2
    * LANDSAT/LE07/C02/T1_L2
    * LANDSAT/LC08/C02/T1_L2
    * LANDSAT/LC09/C02/T1_L2
    """

    def _aux_image(self, mask_shadows: bool = True, mask_cirrus: bool = True, max_cloud_dist: int = 5000) -> ee.Image:
        """
        Retrieve the auxiliary image containing cloud/shadow masks and cloud distance.

        Parameters
        ----------
        mask_shadows: bool, optional
            Whether to mask cloud shadows.
        mask_cirrus: bool, optional
            Whether to mask cirrus clouds.  Valid for Landsat 8-9 images.
        max_cloud_dist: int, optional
            Maximum distance (m) to look for clouds when forming the 'cloud distance' band.  Valid for
            Sentinel-2 images.

        Returns
        -------
        ee.Image
            An Earth Engine image containing *_MASK and CLOUD_DIST bands.
        """
        ee_image = self._ee_image
        qa_pixel = ee_image.select('QA_PIXEL')

        # construct fill mask from Earth Engine mask and QA_PIXEL
        ee_mask = ee_image.select('SR_B.*').mask().reduce(ee.Reducer.allNonZero())
        fill_mask = qa_pixel.bitwiseAnd(1).eq(0).And(ee_mask).rename('FILL_MASK')

        shadow_mask = qa_pixel.bitwiseAnd(0b10000).neq(0).rename('SHADOW_MASK')
        if mask_cirrus:
            cloud_mask = qa_pixel.bitwiseAnd(0b1100).neq(0).rename('CLOUD_MASK')
        else:
            cloud_mask = qa_pixel.bitwiseAnd(0b1000).neq(0).rename('CLOUD_MASK')

        # combine cloud, shadow and fill masks into cloudless mask
        cloud_shadow_mask = (cloud_mask.Or(shadow_mask)) if mask_shadows else cloud_mask
        cloudless_mask = cloud_shadow_mask.Not().And(fill_mask).rename('CLOUDLESS_MASK')

        # copy cloud distance from existing ST_CDIST band (in 10m units), and clip to max_cloud_dist
        cloud_dist = ee_image.select('ST_CDIST').rename('CLOUD_DIST').toUint16()
        cloud_dist = cloud_dist.where(cloud_dist.gt(ee.Image(max_cloud_dist / 10)), max_cloud_dist / 10)

        return ee.Image([fill_mask, cloud_mask, shadow_mask, cloudless_mask, cloud_dist])


class Sentinel2ClImage(CloudMaskedImage):
    """ Base class for cloud/shadow masking of Sentinel-2 TOA and SR images. """

    def _aux_image(
        self, s2_toa: bool = False, mask_cirrus: bool = True, mask_shadows: bool = True,
        mask_method: CloudMaskMethod = CloudMaskMethod.cloud_prob, prob: float = 60, dark: float = 0.15,
        shadow_dist: int = 1000, buffer: int = 50, cdi_thresh: float = None, max_cloud_dist: int = 5000
    ) -> ee.Image:
        """
        Derive cloud, shadow and validity masks for the encapsulated image.

        Adapted from https://github.com/r-earthengine/ee_extra, under Apache 2.0 license.

        Parameters
        ----------
        s2_toa : bool, optional
            S2 TOA/SR collection.  Set to True if this image is from COPERNICUS/S2, or False if it is from
            COPERNICUS/S2_SR.
        mask_cirrus: bool, optional
            Whether to mask cirrus clouds.  Valid for Landsat 8-9 images, and for Sentinel-2 images with
            the `qa` ``mask_method``.
        mask_shadows: bool, optional
            Whether to mask cloud shadows.
        mask_method: CloudMaskMethod, str, optional
            Method used to mask clouds.  Valid for Sentinel-2 images.  See :class:`~geedim.enums.CloudMaskMethod` for
            details.
        prob: float, optional
            Cloud probability threshold (%). Valid for Sentinel-2 images with the `cloud-prob` ``mask-method``.
        dark: float, optional
            NIR threshold [0-1]. NIR values below this threshold are potential cloud shadows.  Valid for Sentinel-2
            images.
        shadow_dist: int, optional
            Maximum distance (m) to look for cloud shadows from cloud edges.  Valid for Sentinel-2 images.
        buffer: int, optional
            Distance (m) to dilate cloud/shadow.  Valid for Sentinel-2 images.
        cdi_thresh: float, optional
            Cloud Displacement Index threshold. Values below this threshold are considered potential clouds.
            If this parameter is not specified (=None), the index is not used.  Valid for Sentinel-2 images.
            See https://developers.google.com/earth-engine/apidocs/ee-algorithms-sentinel2-cdi for details.
        max_cloud_dist: int, optional
            Maximum distance (m) to look for clouds when forming the 'cloud distance' band.  Valid for
            Sentinel-2 images.

        Returns
        -------
        ee.Image
            An Earth Engine image containing *_MASK and CLOUD_DIST bands.
        """
        mask_method = CloudMaskMethod(mask_method)

        def get_cloud_prob(ee_im):
            """Get the cloud probability image from COPERNICUS/S2_CLOUD_PROBABILITY that corresponds to `ee_im`."""
            filt = ee.Filter.eq('system:index', ee_im.get('system:index'))
            cloud_prob = ee.ImageCollection('COPERNICUS/S2_CLOUD_PROBABILITY').filter(filt).first()
            return cloud_prob.rename('CLOUD_PROB')

        def get_cloud_mask(ee_im, cloud_prob=None):
            """Get the cloud mask for ee_im"""
            if mask_method == CloudMaskMethod.cloud_prob:
                if not cloud_prob:
                    cloud_prob = get_cloud_prob(ee_im)
                cloud_mask = cloud_prob.gte(prob)
            else:
                qa = ee_im.select('QA60')
                cloud_mask = qa.bitwiseAnd(1 << 10).neq(0)
                if mask_cirrus:
                    cloud_mask = cloud_mask.Or(qa.bitwiseAnd(1 << 11).neq(0))
            return cloud_mask.rename('CLOUD_MASK')

        def get_cdi_cloud_mask(ee_im):
            """
            Get a CDI cloud mask for ee_im.
            See https://developers.google.com/earth-engine/apidocs/ee-algorithms-sentinel2-cdi for more detail.
            """
            if s2_toa:
                s2_toa_image = ee_im
            else:
                # get the Sentinel-2 TOA image that corresponds to ee_im
                idx = ee_im.get('system:index')
                s2_toa_image = (ee.ImageCollection('COPERNICUS/S2').filter(ee.Filter.eq('system:index', idx)).first())
            cdi_image = ee.Algorithms.Sentinel2.CDI(s2_toa_image)
            return cdi_image.lt(cdi_thresh).rename('CDI_CLOUD_MASK')

        def get_shadow_mask(ee_im, cloud_mask):
            """Given a cloud mask, get a shadow mask for ee_im."""
            dark_mask = ee_im.select('B8').lt(dark * 1e4)
            if not s2_toa:
                dark_mask = ee_im.select('SCL').neq(6).And(dark_mask)

            proj = get_projection(ee_im, min_scale=False)
            # Note:
            # S2 MEAN_SOLAR_AZIMUTH_ANGLE (SAA) appears to be measured clockwise with 0 at N (i.e. shadow goes in the
            # opposite direction), directionalDistanceTransform() angle appears to be measured clockwise with 0 at W.
            # So we need to add/subtract 180 to SAA to get shadow angle in S2 convention, then add 90 to get
            # directionalDistanceTransform() convention i.e. we need to add -180 + 90 = -90 to the SAA.  This is not
            # the same as in the EE tutorial which is 90-SAA
            # (https://developers.google.com/earth-engine/tutorials/community/sentinel-2-s2cloudless).
            shadow_azimuth = ee.Number(-90).add(ee.Number(ee_im.get('MEAN_SOLAR_AZIMUTH_ANGLE')))
            proj_pixels = ee.Number(shadow_dist).divide(proj.nominalScale()).round()
            # Project the cloud mask in the direction of the shadows it will cast.
            cloud_cast_proj = cloud_mask.directionalDistanceTransform(shadow_azimuth, proj_pixels).select('distance')
            # The reproject is necessary to force calculation at the correct scale - the coarse proj scale is used to
            # improve processing times.
            cloud_cast_mask = cloud_cast_proj.mask().reproject(crs=proj.crs(), scale=proj.nominalScale())
            return cloud_cast_mask.And(dark_mask).rename('SHADOW_MASK')

        # gather and combine the various masks
        ee_image = self.ee_image
        cloud_prob = get_cloud_prob(ee_image) if mask_method == CloudMaskMethod.cloud_prob else None
        cloud_mask = get_cloud_mask(ee_image, cloud_prob=cloud_prob)
        if cdi_thresh is not None:
            cloud_mask = cloud_mask.And(get_cdi_cloud_mask(ee_image))
        if mask_shadows:
            shadow_mask = get_shadow_mask(ee_image, cloud_mask)
            cloud_shadow_mask = cloud_mask.Or(shadow_mask)
        else:
            cloud_shadow_mask = cloud_mask

        # do a morphological opening type operation that removes small (20m) blobs from the mask and then dilates
        cloud_shadow_mask = cloud_shadow_mask.focal_min(20, units='meters').focal_max(buffer, units='meters')
        # derive a fill mask from the Earth Engine mask for the surface reflectance bands
        fill_mask = ee_image.select('B.*').mask().reduce(ee.Reducer.allNonZero())
        # Clip this mask to the image footprint.  (Without this step we get memory limit errors on download.)
        fill_mask = fill_mask.clip(ee_image.geometry()).rename('FILL_MASK')

        # combine all masks into cloudless_mask
        cloudless_mask = (cloud_shadow_mask.Not()).And(fill_mask).rename('CLOUDLESS_MASK')

        # construct and return the auxiliary image
        aux_bands = [fill_mask, cloud_mask, cloudless_mask]
        if mask_shadows:
            aux_bands.append(shadow_mask)
        if mask_method == CloudMaskMethod.cloud_prob:
            aux_bands.append(cloud_prob)

        cloud_dist = self._cloud_dist(cloudless_mask=cloudless_mask, max_cloud_dist=max_cloud_dist)
        return ee.Image(aux_bands + [cloud_dist])


class Sentinel2SrClImage(Sentinel2ClImage):
    """ Class for cloud/shadow masking of Sentinel-2 SR (COPERNICUS/S2_SR) images. """

    def _aux_image(self, s2_toa: bool = False, **kwargs) -> ee.Image:
        return Sentinel2ClImage._aux_image(self, s2_toa=False, **kwargs)


class Sentinel2ToaClImage(Sentinel2ClImage):
    """ Class for cloud/shadow masking of Sentinel-2 TOA (COPERNICUS/S2) images. """

    def _aux_image(self, s2_toa: bool = False, **kwargs) -> ee.Image:
        return Sentinel2ClImage._aux_image(self, s2_toa=True, **kwargs)


def class_from_id(image_id: str) -> type:
    """ Return the *Image class that corresponds to the provided Earth Engine image/collection ID. """
    ee_coll_name, _ = split_id(image_id)
    if image_id in geedim.schema.collection_schema:
        return geedim.schema.collection_schema[image_id]['image_type']
    elif ee_coll_name in geedim.schema.collection_schema:
        return geedim.schema.collection_schema[ee_coll_name]['image_type']
    else:
        return MaskedImage
